import pyscratch as pysc
from pyscratch import game

import player,player_bullet, enemy, enemy_bullet

width = 720
height = 1280
sur = pysc.create_rect((50,50,50), width, height)


game.set_backdrops([sur])
game.switch_backdrop(0)

game.update_screen_mode((width, height))
game.start(show_mouse_position=True, event_count=True)
#=game.save_sprite_states()




import pyscratch as pysc
from pyscratch import game
import  trees, player

sur = pysc.create_rect((169,234,126), 1280, 720)
sur_sky = pysc.create_rect((196,241,255), 1280, 200)
sur.blit(sur_sky, (0,0))

game.set_backdrops([sur])
game.switch_backdrop(0)

game.start(show_mouse_position=True, use_frame_time=True)
game.save_sprite_states()




import numpy as np
import pyscratch as pysc
from pyscratch import game


def StandardBullet(position, direction, speed):

    bullet = pysc.create_animated_sprite(
        "assets/used_by_examples/bullet_hell/normal_bullet",
        position=position)
    
    bullet.direction = direction
    bullet.set_scale(1.5)

    
    bullet['movement_event'] = movement_event = bullet.when_timer_reset(1/game.framerate).add_handler(lambda _: bullet.move_indir(speed))
    bullet.when_timer_reset(0.2).add_handler(lambda _: bullet.next_frame())
    
    def detect_collision(_):
        player = game['player']

        if bullet.is_touching(player):
            movement_event.remove()
            bullet.when_timer_reset(0.1, 1).add_handler(lambda _: bullet.remove())

    bullet.when_timer_reset(1/game.framerate).add_handler(detect_collision)
    return bullet
        
game['StandardBullet'] = StandardBullet



def ExplodingBullet(position, direction, lifespan):
    speed = 5
    main_bullet = StandardBullet(position, direction, speed)

    def explode():
        main_bullet['movement_event'].remove()
        main_bullet.remove()
        for i in range(12):
            StandardBullet((main_bullet.x, main_bullet.y), i*30, speed*1.5)

    
    main_bullet.when_timer_reset(lifespan, 1).add_handler(lambda _: explode())


game['ExplodingBullet'] = ExplodingBullet


def Laser(pos0, pos1, duration):

    len_x = pos1[0]-pos0[0]
    len_y = pos1[1]-pos0[1]

    length = (len_x**2 + len_y**2)**(1/2)

    
    path = "assets/used_by_examples/bullet_hell/lasers/3.png"
    img = pysc.load_image(path)
    w = img.get_width()

    scaled_img = pysc.scale_to_fill(img, (w, length))



    laser_sp = pysc.Sprite(dict(always=[scaled_img]))

    laser_sp.direction = -np.arctan(len_x/len_y)/np.pi*180
    laser_sp.x = (pos1[0]+pos0[0])/2
    laser_sp.y = (pos1[1]+pos0[1])/2
    laser_sp.set_transparency(0)
    def change_alpha(_):
        for i in range(30):
            laser_sp.set_transparency(i/30)
            yield 0.01
        for i in range(30):
            laser_sp.set_transparency(1-i/30)
            yield 0.01            
        laser_sp.remove()
    laser_sp.when_timer_above(0).add_handler(change_alpha)


game['Laser'] = Laser


    
    




import pyscratch as pysc
from pyscratch import game

msg = pysc.create_single_costume_sprite("assets/youlose.png")

def gamestart():
    """scale and hide on game start"""
    msg.scale_by(10)
    msg.hide()
    msg.x = 150
    msg.y = 150

msg.when_game_start().add_handler(gamestart)


def show_msg():
    """
    show message when lose
    """
    msg.show()

msg.when_backdrop_switched("lose").add_handler(show_msg)




import pygame


def render_wrapped_file_name(text, max_chars, font, color=(255, 255, 255), bg_color=None, max_lines=None):
    """
    THIS FUNCTION IS WRITTEN BY AN AI

    Renders a file name like a file explorer: wrapping intelligently while preserving the extension.

    :param text: The file name string to render.
    :param max_chars: Maximum number of characters per line (approximate width).
    :param font: Pygame font object to render with.
    :param color: Text color (default white).
    :param bg_color: Background color (default None = transparent).
    :param max_lines: Optional maximum number of lines to render. Extra text will be truncated with '...'.
    :return: A Pygame surface with the rendered multiline text.
    """
    def split_file_name(text):
        if '.' in text and not text.startswith('.') and text.rfind('.') > 0:
            idx = text.rfind('.')
            return text[:idx], text[idx+1:]
        else:
            return text, ''

    def wrap_lines(base, ext, max_chars, max_lines=None):
        words = base.split(' ')
        lines = []
        current_line = ''

        for word in words:
            test_line = (current_line + ' ' + word).strip()
            if len(test_line) <= max_chars:
                current_line = test_line
            else:
                if current_line:
                    lines.append(current_line)
                    if max_lines and len(lines) == max_lines:
                        return lines[:-1] + [truncate_with_ellipsis(current_line, max_chars)]
                # Break long word
                while len(word) > max_chars:
                    lines.append(word[:max_chars])
                    word = word[max_chars:]
                    if max_lines and len(lines) == max_lines:
                        return lines[:-1] + [truncate_with_ellipsis(lines[-1], max_chars)]
                current_line = word

        if current_line:
            lines.append(current_line)
            if max_lines and len(lines) > max_lines:
                lines = lines[:max_lines]
                lines[-1] = truncate_with_ellipsis(lines[-1], max_chars)

        # Add extension
        if ext:
            if len(lines[-1] + '.' + ext) <= max_chars:
                lines[-1] += '.' + ext
            elif not max_lines or len(lines) < max_lines:
                lines.append('.' + ext)
            else:
                lines[-1] = truncate_with_ellipsis(lines[-1], max_chars)

        return lines

    def truncate_with_ellipsis(line, max_chars):
        return line[:max(0, max_chars - 3)] + '...'

    # Prepare
    base, ext = split_file_name(text)
    lines = wrap_lines(base, ext, max_chars, max_lines)

    # Render
    line_height = font.get_linesize()
    surface_height = line_height * len(lines)
    surface_width = max(font.size(line)[0] for line in lines)

    rendered_surface = pygame.Surface((surface_width, surface_height), pygame.SRCALPHA)
    if bg_color:
        rendered_surface.fill(bg_color)

    for i, line in enumerate(lines):
        text_surface = font.render(line, True, color)
        rendered_surface.blit(text_surface, (0, i * line_height))

    return rendered_surface


import pyscratch as pysc
from pyscratch import game

def Platform(position):
    sp = pysc.create_animated_sprite("assets/used_by_examples/doodle_jump/platform")
    sp.x = position[0]
    sp['world_y'] = position[1]
    game.move_to_back(sp)

    sp.y = sp['world_y'] -game['view_y']
    player = game['player']
    def movement_and_touch(_):
        while True:
            yield 1/game.framerate
            sp.y = sp['world_y'] -game['view_y']
            if sp.is_touching(player) and (player['vy']>0):
                game.broadcast_message('jump')

    sp.when_timer_above(0).add_handler(movement_and_touch)
    return sp

game.when_game_start().add_handler(lambda: Platform((300, 1000)))
game.when_game_start().add_handler(lambda: Platform((300, 800)))


def create_platform():

    platforms = []
    for i in range(20):
        y = i*pysc.random_number(70, 100)
        p = Platform((pysc.random_number(0, game.screen_width), y))
        platforms.append(p)

    while True:
        yield 1/game.framerate
        platforms = [p for p in platforms if p.y<0]
        if len(platforms) < 5:
            #print(len(platforms))

            y = 3*pysc.random_number(70, 100)
            p = Platform((pysc.random_number(0, game.screen_width), game['view_y']-y))
            print(p.y)
            platforms.append(p)

    



    

game.when_game_start().add_handler(create_platform)


"""
The reference manual is intended to provide technical specifications for the functions in this library. 
If you are new to this library, visit [getting-started](../getting-started/). 
# Namespace
Everything in this library are directly under the pyscratch namespace. 
So for example, you can use `pyscratch.game` to refer to `pyscratch.game_module.game` or `pyscratch.Sprite` to refer to `pyscratch.sprite.Sprite` etc. 

When you do `import pyscratch as pysc`, they become just `pysc.game`, `pysc.Sprite` and `pysc.random_number` etc. 


# Main Objects
You will be mainly interacting of these following types of objects: 
## 1. pyscratch.game_module.game (pysc.game)
This is an object instead a type of objects. This object represents the game itself. 
You use it to start the game, to play sound, change backdrops and make global changes to the game. 
You can also use it to create events, but it is more convenient to create the events from the sprite.

## 2. pyscratch.sprite.Sprite 
The Sprite objects are the objects that represents the sprites in the game.


## 3. pyscratch.event.Event & pyscratch.event.Condition
This is the object you will get when you run `my_event = sprite1.when_game_start()`
The only way you will interact with this object is to add handler functions to it and to remove it. 


## 4. Pygame objects that represents the assets
pygame is another library that this library depends on. 
You only need to interact with these objects when you are loading the assets (for the sprites, backdrops and text styles)

### 4.1. pygame.Surface
A pygame object that represents an image. 
All the images in this library are represented as pygame.Surface objects. 
You can use pyscratch.helper.load_image to load an image for you. 

### 4.2. pygame.font.Font
A pygame object that represent the font (or the style of the text). 
Some functions require the font object to render the text.
To create a font object: 
```python
font = pygame.font.SysFont(None, 48)  # None = default font, 48 = font size
```
"""

from .event import *
from .game_module import *
from .helper import *
from .sprite import *




import numpy as np
import pyscratch as pysc
game = pysc.game


colour = 0, 0, 0, 50

cir0 = pysc.create_circle_sprite(colour, 30)
game['c0'] = cir0

cir1 = pysc.create_circle_sprite(colour, 30)
game['c1'] = cir1

# event game start: initial setting
def on_start():

    cir0.set_draggable(True)
    cir1.set_draggable(True)

    # pysc.game.bring_to_front(cir)
    # pysc.game.bring_to_front(cir2)

    cir0.x = game['ss_view_topleft'][0] + 1
    cir0.y = game['ss_view_topleft'][1] + 1
    
    cir1.x = game['ss_view_buttom_right'][0]-1
    cir1.y = game['ss_view_buttom_right'][1]-1


    cir0.oob_limit=np.inf
    cir1.oob_limit=np.inf

game.when_game_start([cir0, cir1]).add_handler(on_start)



# useful functions
def value_if_none(v, default):
    if v is None:
        return default
    return v

def value_if_not(v, default):
    if not v:
        return default
    return v




# event game start: set the cutting offset and the frame size if the sprite is dragged
def on_start2():
    while True:
        yield 1/30
        ss_sprite: pysc.Sprite = game['ss_sprite']
        if not ss_sprite: continue
        game.bring_to_front(cir0)
        game.bring_to_front(cir1)

        mouse_down = pysc.get_mouse_presses()[0]
        if not mouse_down: 
            try:
                cir0.x = game['offset_x']*ss_sprite.scale_factor + ss_sprite.rect.left
                cir1.x = game['limit_x']*ss_sprite.scale_factor + ss_sprite.rect.left        

                cir0.y = game['offset_y']*ss_sprite.scale_factor + ss_sprite.rect.top
                cir1.y = game['limit_y']*ss_sprite.scale_factor + ss_sprite.rect.top    
            except:
                pass          
            continue
        if not (cir0.is_touching_mouse() or cir1.is_touching_mouse()): continue



        # top left of the selected rect
        cx0 = min(cir0.x, cir1.x)
        cy0 = min(cir0.y, cir1.y)
        
        # bottom right of the selected rect
        cx1 = max(cir0.x, cir1.x)
        cy1 = max(cir0.y, cir1.y)        

        # top left of the sprite sheet
        x0, y0 = ss_sprite.rect.topleft

        img_w, img_h = ss_sprite['original_width'], ss_sprite['original_height']

        # scale factor
        scale_factor = ss_sprite.scale_factor
        
        # RC
        n_col = value_if_not(game['n_col'], 1)
        n_row = value_if_not(game['n_row'], 1)

        # O
        game['offset_x'] = Ox = max(np.floor((cx0-x0)/scale_factor), 0)
        game['offset_y'] = Oy = max(np.floor((cy0-y0)/scale_factor), 0)

        # L
        game['limit_x'] = Lx = min(np.ceil((cx1-x0)/scale_factor), img_w-1)
        game['limit_y'] = Ly = min(np.ceil((cy1-y0)/scale_factor), img_h-1) 

        # update the size
        game['size_x'] = Lx - Ox + 1
        game['size_y'] = Ly - Oy + 1

        # pix
        game['pixel_x'] = np.floor(game['size_x']/n_col)
        game['pixel_y'] = np.floor(game['size_y']/n_row)





pysc.game.when_game_start([cir0, cir1]).add_handler(on_start2)


def on_offset_x_change(data):

    if not (size_x := game['size_x']): return 
    if not (offset_x := game['offset_x']): 
        game['offset_x'] = 0
        offset_x = 0         

    ss_sprite: pysc.Sprite = game['ss_sprite']
    if not ss_sprite: return 
    
    game['limit_x'] = min(size_x+offset_x-1, ss_sprite['original_width']-1)

    game['size_x'] =  game['limit_x'] - offset_x + 1

    cir0.x = offset_x*ss_sprite.scale_factor + ss_sprite.rect.left
    cir1.x = game['limit_x']*ss_sprite.scale_factor + ss_sprite.rect.left

    game['n_col'] = np.floor(game['size_x']/game['pixel_x'])

game.when_receive_message('offset_x_change').add_handler(on_offset_x_change)

def on_offset_y_change(data):

    if not (size_y := game['size_y']): return 
    if (offset_y := game['offset_y']) is None: 
        game['offset_y'] = 0
        offset_y = 0
         
    ss_sprite: pysc.Sprite = game['ss_sprite']
    if not ss_sprite: return 
    
    game['limit_y'] = min(size_y+offset_y-1, ss_sprite['original_height']-1)

    game['size_y'] =  game['limit_y'] - offset_y + 1

    cir0.y = offset_y*ss_sprite.scale_factor + ss_sprite.rect.top
    cir1.y = game['limit_y']*ss_sprite.scale_factor + ss_sprite.rect.top

    game['n_row'] = np.floor(game['size_y']/game['pixel_y'])

game.when_receive_message('offset_y_change').add_handler(on_offset_y_change)


def on_n_row_change(data):
    if not game['size_y']: return 
    game['pixel_y'] = np.floor(game['size_y']/value_if_not(game['n_row'], 1))

game.when_receive_message('n_row_change').add_handler(on_n_row_change)


def on_n_col_change(data):
    if not game['size_x']: return 
    game['pixel_x'] = np.floor(game['size_x']/value_if_not(game['n_col'], 1))

game.when_receive_message('n_col_change').add_handler(on_n_col_change)



def on_pix_x_change(data):
    if not game['size_x']: return 
    game['n_col'] = np.floor(game['size_x']/value_if_not(game['pixel_x'], 1))


game.when_receive_message('pixel_x_change').add_handler(on_pix_x_change)

def on_pix_y_change(data):
    if not game['size_y']: return 
    game['n_row'] = np.floor(game['size_y']/value_if_not(game['pixel_y'], 1))


game.when_receive_message('pixel_y_change').add_handler(on_pix_y_change)



def on_scale_change(data):
    ss_sprite: pysc.Sprite = game['ss_sprite']
    if game['offset_x'] is None: return 
    if game['limit_x'] is None: return 
    if game['offset_y'] is None: return 
    if game['limit_y'] is None: return 
    

    cir0.x = game['offset_x']*ss_sprite.scale_factor + ss_sprite.rect.left
    cir1.x = game['limit_x']*ss_sprite.scale_factor + ss_sprite.rect.left

    cir0.y = game['offset_y']*ss_sprite.scale_factor + ss_sprite.rect.top
    cir1.y = game['limit_y']*ss_sprite.scale_factor + ss_sprite.rect.top


game.when_receive_message('ss_sprite_scale_change').add_handler(on_pix_y_change)



import pyscratch as pysc
from pyscratch import game

friend = pysc.create_single_costume_sprite("assets/fish_orange_outline.png")

def movement():
    """
    when_game_start: 
    the movement of the friend: always moving away from the curosr
    """
    centre = game.screen_width/2, game.screen_height/2
    while True:
        yield 1/game.framerate
        mouse_x, mouse_y = pysc.get_mouse_pos()
        if friend.distance_to((mouse_x, mouse_y)) < 200:
            friend.point_towards_mouse()
            friend.direction += 180
            friend.move_indir(2)
        
        else:
            friend.point_towards(centre)
            friend.move_indir(2)

friend.when_game_start().add_handler(movement)
        


def check_if_centred():
    """
    when_game_start: 
    move away when near the centre
    """
    centre = game.screen_width/2, game.screen_height/2
    while True:
        yield 1/game.framerate
        if friend.distance_to(centre) < 50:
            game['score'] += 10

            new_x = pysc.random_number(0, game.screen_width)
            new_y = pysc.random_number(0, game.screen_height)

            friend.x = new_x
            friend.y = new_y
    
friend.when_game_start().add_handler(check_if_centred)
        

def when_clicked():
    """
    flash the friend when clicked
    """
    for i in range(5):
        friend.hide()
        yield 0.1
        friend.show()
        yield 0.1

friend.when_this_sprite_clicked().add_handler(when_clicked)


def lose():
    """
    hide when lose
    """
    friend.hide()

friend.when_backdrop_switched("lose").add_handler(lose)


"""
shared_data
    animation_dict (order -> animation selection button)
    sprite_folder_path
    path (read): navigation path
    main_bottom_size (read)
    main_bottom_panel (read)

broadcast messages:
    folder_update (navigation path)
    change_sprite_selection (path of the sprite folder)
    change_animation (path of the animation folder)


receive messages:
    change_sprite_selection
        change the animation panel when another sprite is selected
"""


from pathlib import Path
from typing import cast
import pyscratch as pysc
from settings import *

container_width = SCREEN_WIDTH - LEFT_PANEL_WIDTH - RIGHT_PANEL_WIDTH - PANEL_MARGIN*3
container_height = 300

# container
container = pysc.create_rect_sprite((230, 230, 230), container_width, container_height)
container.x = container_width/2+LEFT_PANEL_WIDTH+PANEL_MARGIN
container.y = SCREEN_HEIGHT-container_height/2-PANEL_MARGIN
pysc.game['main_bottom_panel'] = container
pysc.game['main_bottom_size'] = container_width, container_height

# sprite function
def NewButton(text):
    button_w, button_h = 150, 50
    button = pysc.create_rect_sprite((221, 221, 221), button_w, button_h)
    button.write_text(text, DEFAULT_FONT48, colour=(0,0,0), offset=(button_w/2, button_h/2))
    button.lock_to(container, (-(container_width-button_w)/2, -(container_height-button_h)/2), reset_xy=True)
    return button


# sprite: add sprite button
add_sprite_button = NewButton("New Sprite")
add_sprite_button.y=10
add_sprite_button.x=200

# event: the click of the add sprite button
def on_click1():
    folder_path: Path = pysc.game.shared_data['path']
    
    c = 0
    while True:
        
        new_folder = (folder_path / f"unnamed_sprite_{c}")
        if not new_folder.exists():
            new_folder.mkdir()
            break
        else:
            c+=1
    pysc.game.broadcast_message('folder_update', folder_path)
    pysc.game.broadcast_message('cut_or_nav_mode_change', 'nav')
    pysc.game.broadcast_message('change_sprite_selection', new_folder)
        
add_sprite_button.when_this_sprite_clicked().add_handler(on_click1)


# sprite function: animation selection
def AnimationSelection(order, path):
    

    w, h = 120, 30
    on_select = pysc.create_rect((150,150, 150), w, h)
    not_on_select = pysc.create_rect((127, 127, 127), w, h)
    sprite = pysc.Sprite(dict(on_select=[on_select], not_on_select=[not_on_select]))

    sprite.lock_to(container, (-(container_width-w)/2, -(container_height-h)/2))
    sprite.x = 3
    sprite.y = order*(h+3)+3

    sprite.set_animation('on_select')
    sprite.write_text(path.stem, DEFAULT_FONT24, offset=(w/2, h/2))

    sprite.set_animation('not_on_select')
    sprite.write_text(path.stem, DEFAULT_FONT24, offset=(w/2, h/2))



    def on_click():
        sprite.set_animation('on_select')

        pysc.game.broadcast_message('change_animation', path)
        pysc.game.broadcast_message('deselect', sprite)


    sprite.when_this_sprite_clicked().add_handler(on_click)


    def on_deselect(s):
        if not sprite is s:
            sprite.set_animation('not_on_select')

    sprite.when_receive_message('deselect').add_handler(on_deselect)

    return sprite

# event: msg: change_sprite_selection: 
# remove and create the animation buttons when sprite change
pysc.game['animation_dict'] = {}
def on_msg_change_sprite_selection(sprite_folder:Path):
    """
    change the animation selection buttons
    """

    pysc.game['sprite_folder_path'] = sprite_folder

    #pysc.game['frame_dict'] = 



    # remove all the old buttons
    for k, v in pysc.game['animation_dict'].items():
        v.remove()
    pysc.game['animation_dict'] = {}

    animation_dict = pysc.game['animation_dict']
    

    # add the new buttons
    c = 0
    for f in sprite_folder.iterdir():
        if f.is_dir():
            animation_dict[c] = AnimationSelection(c, f)
            c+=1

pysc.game.when_receive_message("change_sprite_selection").add_handler(on_msg_change_sprite_selection)




# sprite: add animation button
add_animation_button = NewButton("New Animation")
add_animation_button.x = 370
add_animation_button.y = 10

# event: click of the add animation button  
def on_click2():


    if not (folder_path:= pysc.game.shared_data.get('sprite_folder_path')):
        pysc.game.broadcast_message('warning', "Sprite folder not selected yet")
        return
    
    folder_path = cast(Path, folder_path)
    c = 0
    while True:
        
        
        new_folder = (folder_path / f"animation_{c}")
        if not new_folder.exists():
            new_folder.mkdir()
            break
        else:
            c+=1

    pysc.game.broadcast_message('folder_update', pysc.game.shared_data['path'])
    
    # no actual change. just to trigger the update
    pysc.game.broadcast_message('change_sprite_selection', folder_path)

        
add_animation_button.when_this_sprite_clicked().add_handler(on_click2)


"""
shared_data:
    display_sprite
    scale_factor
    frame_interval
    is_playing

receive messages:
    preview_click
        change the frame of the display sprite when the preview is clicked

    scale_factor_change
        change the scale of the display sprite    
    
    change_animation_done
        destory and create the display sprite for a new animation
"""

from pathlib import Path
from typing import Dict

from pygame import Surface
import pyscratch as pysc
from settings import *


pysc.game['display_sprite'] = None


def create_display_sprite(frames):

    #frame_list = [c['surface'] for c in pysc.game['frame_card_list']]
    #print(frames)
    if display_sprite:=pysc.game['display_sprite']:
        display_sprite.remove()
        x, y = display_sprite.x, display_sprite.y
    else:
        x, y = 350, 200

    if not len(frames):
        pysc.game['display_sprite'] = None
        return


    sprite = pysc.Sprite({'always':frames})
    sprite.set_xy((x, y))
    
    sprite.set_draggable(True)

    if pysc.game['scale_factor']:
        sprite.set_scale(pysc.game['scale_factor'])

    pysc.game['display_sprite'] = sprite

    def play(_):
        #TODO: if the user put a big value in frame_interval, and change it back to a small one
        # the animation will wait for the big frame_interval until it comes back again 
        while True:
            itv = pysc.game['frame_interval']
            if pysc.game['is_playing'] and itv:
                sprite.next_frame()
            yield itv if itv else 0.1

    sprite.when_timer_above(0).add_handler(play)


    def on_scale_factor_change(scale):
        print(scale)
        if scale: 
            sprite.set_scale(scale)

    sprite.when_receive_message("scale_factor_change").add_handler(on_scale_factor_change)
    sprite.when_receive_message('preview_click').add_handler(lambda d: sprite.set_frame(d['order']))



pysc.game.when_receive_message("change_animation_done").add_handler(create_display_sprite)


from pathlib import Path
from typing import List, Tuple, cast
import pyscratch as pysc
from pyscratch.sprite import Sprite
from settings import *
from .file_display import FileDisplay

width, height = RIGHT_PANEL_WIDTH, SCREEN_HEIGHT - PANEL_MARGIN - BOTTOM_RIGHT_PANEL_HEIGHT

colour = 255, 255, 255
folder_navigation = pysc.create_rect_sprite(colour, width, height)

folder_navigation.set_xy((SCREEN_WIDTH - PANEL_MARGIN - RIGHT_PANEL_WIDTH/2,  SCREEN_HEIGHT/2-BOTTOM_RIGHT_PANEL_HEIGHT/2))

topleft = folder_navigation.x-width/2, folder_navigation.y-height/2

folder_navigation['file_display_list'] = []

def on_msg_folder_update(path: Path):
    #folder_navigation.sprite_data['path'] = path
    pysc.game.shared_data['path'] = path

    for fdisp in folder_navigation['file_display_list']:
        fdisp.remove()

    folder_navigation['file_display_list']  = []
    file_display_list = folder_navigation['file_display_list'] 

    c = 0
    for f in path.iterdir():
        fdisp = FileDisplay(f, c, topleft)
        if fdisp: 
            c+=1
            file_display_list.append(fdisp)

folder_navigation.when_receive_message('folder_update').add_handler(on_msg_folder_update)


def on_msg_back_nav(_):
    path: Path = pysc.game['path']
    pysc.game.broadcast_message('folder_update', path.parent)


    
folder_navigation.when_receive_message('back_nav').add_handler(on_msg_back_nav)



def on_msg_mode_change(mode):
    if mode == 'nav':
        folder_navigation.show()
        pysc.game.move_to_back(folder_navigation)
    else:
        folder_navigation.hide()

folder_navigation.when_receive_message('cut_or_nav_mode_change').add_handler(on_msg_mode_change)
pysc.game['folder_navigation'] = folder_navigation


import pyscratch as pysc
game = pysc.game
target = pysc.create_single_costume_sprite("assets/target_b.png")

# 3. Flow control
def follow_mouse():
    game.bring_to_front(target)
    while True:
        yield 1/game.framerate
        mouse_x, mouse_y = pysc.get_mouse_pos()
        target.x = mouse_x
        target.y = mouse_y
    
target.when_game_start().add_handler(follow_mouse)


"""
Everything in this module is directly under the pyscratch namespace. 
For example, instead of doing `pysc.game_module.is_key_pressed`,
you can also directly do `pysc.is_key_pressed`.
"""

from __future__ import annotations
from functools import cache
from os import PathLike
import os
from pathlib import Path
import threading
import time
import json, inspect
from typing_extensions import deprecated

import numpy as np
import pygame
import pymunk
from .event import _ConditionInterface, Event, Condition, TimerCondition, _declare_callback_type
from pymunk.pygame_util import DrawOptions
from typing import Any, Callable, Generic, Iterable, Literal, Optional, List, Dict, ParamSpec, Set, Tuple, TypeVar, Union, cast
from typing import TYPE_CHECKING
from . import helper

if TYPE_CHECKING:
    from .sprite import Sprite

def _collision_begin(arbiter, space, data):
    game = cast(Game, data['game'])
    game._contact_pairs_set.add(arbiter.shapes) 

    for e, (a,b) in game._trigger_to_collision_pairs.items():
        if (a._shape in arbiter.shapes) and (b._shape in arbiter.shapes):
            e.trigger(arbiter)


    colliding_types = arbiter.shapes[0].collision_type, arbiter.shapes[1].collision_type
    collision_allowed = True
    for collision_type, (allowed, triggers) in game._collision_type_to_trigger.items():
        if collision_type in colliding_types: 
            [t.trigger(arbiter) for t in triggers]
            collision_allowed = collision_allowed and allowed
        

    if (arbiter.shapes[0].collision_type == 0) or (arbiter.shapes[1].collision_type == 0):
        collision_allowed = False


    return collision_allowed

def _collision_separate(arbiter, space, data):
    game = cast(Game, data['game'])

    if arbiter.shapes in game._contact_pairs_set:
        game._contact_pairs_set.remove(arbiter.shapes)


    reverse_order = arbiter.shapes[1], arbiter.shapes[0]
    if reverse_order in game._contact_pairs_set:
        game._contact_pairs_set.remove(reverse_order)


class _CloneEventManager:

    def __init__(self):
        # TODO: removed sprites stay here forever
        self.identical_sprites_and_triggers: List[Tuple[Set[Sprite], List[Event]]] = []

    def new_trigger(self, sprite:Sprite, trigger:Event):
        new_lineage = True
        for identical_sprites, triggers in self.identical_sprites_and_triggers:
            if sprite in identical_sprites: 
                new_lineage = False
                triggers.append(trigger)

        if new_lineage: 
            self.identical_sprites_and_triggers.append((set([sprite]), [trigger]))
                

    def on_clone(self, old_sprite:Sprite, new_sprite:Sprite):
        # so that the cloning of the cloned sprite will trigger the same event
        for identical_sprites, triggers in self.identical_sprites_and_triggers:
            if not old_sprite in identical_sprites:
                continue
            identical_sprites.add(new_sprite)

            for t in triggers:
                t.trigger(new_sprite)

        

class _SpriteEventDependencyManager:

    def __init__(self):

        self.sprites: Dict[Sprite, List[Union[Event, _ConditionInterface]]] = {}

    def add_event(self, event: Union[_ConditionInterface, Event], sprites: Iterable[Sprite]):
        """
        TODO: if the event is dependent to multiple sprites, the event will not be
        completely dereferenced until all the sprites on which it depends are removed

        """
        for s in sprites:
            if not s in self.sprites:
                self.sprites[s] = []
            self.sprites[s].append(event)


    def sprite_removal(self, sprite: Sprite):
        
        to_remove = self.sprites.get(sprite)
        if not to_remove:
            return 
    
        for e in to_remove:
            e.remove()

T = TypeVar('T')
"""@private"""
P = ParamSpec('P')
"""@private"""

class _SpecificEventEmitter(Generic[P]):

    def __init__(self):
        self.key2triggers: Dict[Any, List[Event[P]]] = {}

    def add_event(self, key, trigger:Event[P]):
        if not key in self.key2triggers: 
            self.key2triggers[key] = []
        self.key2triggers[key].append(trigger)
        

    def on_event(self, key, *args: P.args, **kwargs: P.kwargs):
        if not key in self.key2triggers: 
            return
        for t in self.key2triggers[key]:
            t.trigger(*args, **kwargs)



class _SavedSpriteStateManager:
    default_filename = "saved_sprite_states.json"
    def __init__(self):
        self.states: Dict[str, Dict[str, Any]]  = {}


    def save_sprite_states(self, all_sprite: Iterable[Sprite], filename=None):
        """
        Save the x, y & direction of the sprites. 

        Usage: 
        ```python
        # main.py
        from pyscratch import game

        game.start() # when you close the game window, the game.start() function finishes.
        game.save_sprite_states() # then this function will be run.   
        
        ```
        """
        loc = {}
        for s in all_sprite:
            loc[s.identifier] = dict(x=s.x, y=s.y, direction=s.direction)

        if not filename:
            filename = self.default_filename
        #caller_file = Path(inspect.stack()[-1].filename)

        json.dump(loc, open(filename, "w"))
        print("Sprite states saved.")
    
    def load_saved_state(self, filename=None):

        if not filename:
            filename = self.default_filename
            
        filename = Path(filename)
        if filename.exists():
            self.states= json.load(open(filename, "r"))


    def get_state_of(self, sprite_id):

        return self.states.get(sprite_id)
    

class Game:
    """
    This is the class that the `game` object belongs to. You cannot create another Game object. 
    To exit the game, either close the window, or to press the escape key (esc) by default
    """
    
    _singleton_lock = False
    def __init__(self):
        """@private"""
        pygame.init()

        assert not Game._singleton_lock, "Already instantiated."
        Game._singleton_lock = True

        # the screen is needed to load the images. 
        self._screen: pygame.Surface  = pygame.display.set_mode((1280, 720), vsync=1)

        self._space: pymunk.Space = pymunk.Space()


        self._draw_options = DrawOptions(self._screen)

        # sounds
        self._mixer = pygame.mixer.init()

        self._sounds = {}

        # shared variables 
        self.shared_data: Dict[Any, Any] = {}
        """
        A dictionary of variables shared across the entire game. You can put anything in it.
        
        The access of the items can be done directly through the game object. 
        For example, `game['my_data'] = "hello"` is just an alias of `game.shared_data['my_data'] = "hello"` 
        
        Bare in mind that the order of executions of different events and different files is arbitrary. 
        Therefore if variable is defined in one event and accessed in another, 
        a KeyError may be raised because variables are only accessible after the definition. 

        Instead, the all the variable should be defined outside the event (before the game start),
        and variables should be accessed only within events to guarantee its definition. 


        Example:
        ```python
        from pyscratch import game 

        # same as `game.shared_data['score_left'] = 0`
        game['score_left'] = 0
        game['score_right'] = 0

        def on_score(side):
            if side == "left":
                game['score_left'] += 1
            else:
                game['score_right'] += 1

            print(f"Left score: {game['score_left']}") 
            print(f"Right score: {game['score_right']}") 

        game.when_received_message('score').add_handler(on_score)
        game.broadcast_message('on_score', 'left')
        ```
        """
        
        # sprite event dependency manager
        self._sprite_event_dependency_manager = _SpriteEventDependencyManager()

        # 
        self._clone_event_manager = _CloneEventManager()
        """@private"""

        # collision detection
        self._trigger_to_collision_pairs: Dict[Event, Tuple[Sprite, Sprite]] = {}

        self._collision_type_pair_to_trigger: Dict[Tuple[int, int], List[Event]] = {}

        self._collision_type_to_trigger: Dict[int, Tuple[bool, List[Event]]] = {}

        self._contact_pairs_set: Set[Tuple[pymunk.Shape, pymunk.Shape]] = set() 


        self._collision_handler = self._space.add_default_collision_handler()

        self._collision_handler.data['game'] = self
        self._collision_handler.begin = _collision_begin
        self._collision_handler.separate = _collision_separate
        
        # sprites updating and drawing
        self._all_sprites = pygame.sprite.Group()

        self._all_sprites_to_show = pygame.sprite.LayeredUpdates(default_layer=1)


        # # scheduled jobs
        # self.pre_scheduled_jobs = []
        # self.scheduled_jobs = []


        self._all_pygame_events = []

        self._all_triggers: List[Event] = [] # these are to be executed every iteration

        self._all_conditions: List[_ConditionInterface] = [] # these are to be checked every iteration

        #self.all_forever_jobs: List[Callable[[], None]] = []
        self._all_message_subscriptions: Dict[str, List[Event]] = {}

        # key events 
        key_event = self.create_pygame_event([pygame.KEYDOWN, pygame.KEYUP])
        key_event.add_handler(self.__key_event_handler)
        self._all_simple_key_triggers: List[Event] = [] # these are to be triggered by self.__key_event_handler only

        # mouse dragging event
        self._dragged_sprite = None
        self._drag_offset = 0, 0
        self.__clicked_sprite =  None
        self._sprite_click_release_trigger:Dict[Sprite, List[Event]] = {}  #TODO: need to be able to destory the trigger here when the sprite is destoryed

        self._sprite_click_trigger:Dict[Sprite, List[Event]] = {}  #TODO: need to be able to destory the trigger here when the sprite is destoryed
        self._mouse_drag_trigger = self.create_pygame_event([pygame.MOUSEBUTTONDOWN, pygame.MOUSEBUTTONUP, pygame.MOUSEMOTION])
        self._mouse_drag_trigger.add_handler(self.__mouse_drag_handler)

        # Backdrops
        self.backdrops_by_key: Dict[str, pygame.Surface] = {}
        self.backdrop_indices_by_key: Dict[str, int] = {}
        self.backdrop_keys: List[str] = []

# TODO: testing needed

        self.__screen_width: int = 0
        self.__screen_height: int = 0
        self.__framerate: float = 0

        # TODO index and key need to be changed together 
        # the index is remained for the next_backdrop function
        self.__backdrop_index = None
        self.__backdrop_key = None

        # TODO: the event is now triggered by the name not the index
        self._backdrop_change_triggers: List[Event] = []

        self._top_edge: Sprite
        self._left_edge: Sprite
        self._bottom_edge: Sprite
        self._right_edge: Sprite


        ## start event
        self._game_start_triggers: List[Event] = []

        ## global timer event
        self._global_timer_triggers: List[Event] = []

    
        self._current_time_ms: float = 0

        self._specific_key_event_emitter: _SpecificEventEmitter[str] = _SpecificEventEmitter()

        self._specific_backdrop_event_emitter: _SpecificEventEmitter[[]] = _SpecificEventEmitter()

        self.max_number_sprite = 1000
        """The maximum number of sprites in the game. Adding more sprites will lead to an error to prevent freezing. Default to 1000."""
        
        self._sprite_count_per_file: Dict[str, int] = {}

        self.__start = False
        """set to false to end the loop"""


        self.update_screen_mode()

        self._saved_states_manager = _SavedSpriteStateManager()
        self.__sprite_last_clicked_for_removal: Optional[Sprite] = None
        self.__started_interactive = False


    def __key_event_handler(self, e):
        up_or_down = 'down' if e.type == pygame.KEYDOWN else 'up'
        keyname = pygame.key.name(e.key)

        self._specific_key_event_emitter.on_event(keyname, up_or_down)

        for t in self._all_simple_key_triggers:
            t.trigger(keyname, up_or_down)

    def __getitem__(self, key):
        return self.shared_data[key]
    
    def __setitem__(self, k, v):
        self.shared_data[k] = v


    def __mouse_drag_handler(self, e):

        if e.type == pygame.MOUSEBUTTONDOWN and e.button == 1: 

            for s in reversed(list(self._all_sprites_to_show)):
                if TYPE_CHECKING:
                    s = cast(Sprite, s)
                    
                # click is on the top sprite only
                if s.is_touching_mouse():
                    self.__clicked_sprite = s
                    self.__sprite_last_clicked_for_removal = s

                    for t in self._sprite_click_trigger[s]:
                        t.trigger()

                    if not s.draggable:
                        break
                    
                    s._set_is_dragging (True)
                    self._dragged_sprite = s
                    offset_x = s._body.position[0]  - e.pos[0]
                    offset_y = s._body.position[1]  - e.pos[1]
                    self._drag_offset = offset_x, offset_y
                    break 



        elif e.type == pygame.MOUSEBUTTONUP  and e.button == 1:
            if self._dragged_sprite: 
                self._dragged_sprite._set_is_dragging(False)
                self._dragged_sprite = None

            # TODO: what happens here?? why AND?
            if self.__clicked_sprite and (temp:= self._sprite_click_release_trigger.get(self.__clicked_sprite)):
                #temp  =  self._sprite_click_release_trigger.get(self.__clicked_sprite)
                #if temp: 
                for t in temp:
                    t.trigger()
                self.__clicked_sprite = None

        elif e.type == pygame.MOUSEMOTION and self._dragged_sprite:
            x = e.pos[0] + self._drag_offset[0]
            y = e.pos[1] + self._drag_offset[1]
            self._dragged_sprite.set_xy((x,y))



    def update_screen_mode(self, *arg, **kwargs):
        """
        Update the screen, taking the arguments for 
        [`pygame.display.set_mode`](https://www.pygame.org/docs/ref/display.html#pygame.display.set_mode).
        

        Use this method to change the screen size:

        `game.update_screen_mode((SCREEN_WIDTH, SCREEN_HEIGHT))`
        """
        self.__screen_args = arg
        self.__screen_kwargs = kwargs
        
    @property
    def screen_width(self):
        """The width of the screen. Not available until the game is started (should not be referenced outside events)."""
        return self.__screen_width
    
    @property
    def screen_height(self):
        """The height of the screen. Not available until the game is started (should not referenced outside events)."""
        return self.__screen_height
    
    @property
    def framerate(self):
        """The frame rate of the game. Not available until the game is started"""
        return self.__framerate
    
    def _do_autoremove(self):
        for s in self._all_sprites:
            if ((s.x < -s.oob_limit) or 
             (s.x > (s.oob_limit + self.__screen_width)) or 
             (s.y < -s.oob_limit) or 
             (s.y > (s.oob_limit + self.__screen_height)) 
            ):     
                #print(s.x, s.y)
                #print(s.oob_limit + self.__screen_width, s.oob_limit + self.__screen_height)
                s.remove()
                print(f"{s} is removed for going out of boundary above the specified limit.")

    def _check_alive(self):

        last_frame_time = 0
        while True:

            for i in range(10):
                time.sleep(.2)
                if not self.__start:
                    return 
            if not self._current_time_ms > last_frame_time:
                print('Stucked in the same frame for more than 1 second. This can happen when an error occur or you are in a infinite loop without yielding. ')
                os._exit(1)

            last_frame_time = self._current_time_ms

    def start_interactive(self, *args, sprite_removal_key='backspace', **kwargs):
        """
        @private
        Start the game on another thread for jupyter notebook (experimental)
        """ 
        if self.__started_interactive: 
            raise RuntimeError("The game must not be restarted. Please restart the kernal. ")
        self.__started_interactive = True    
        
        def remove_sprite(_):
            if s:=self.__clicked_sprite: 
                s.remove()

        self.when_key_pressed(sprite_removal_key).add_handler(remove_sprite)

        t = threading.Thread(target=self.start, args=args, kwargs=kwargs)
        t.start()
        self.__screen_thread = t
        
        
        return t
    
    def stop(self):
        """
        @private
        """
        self.__start = False
        if self.__started_interactive:
            #self.__started_interactive = False
            pygame.display.quit()
        

    def start(
            self, 
            framerate=30, 
            sim_step_min=300, 
            debug_draw=False, 
            event_count=False, 
            show_mouse_position: Optional[bool]=None, 
            exit_key: Optional[str]="escape",
            saved_state_file=None, 
            print_fps = False,
            use_frame_time = False
        ):
        """
        Start the game. 

        Parameters
        ---
        framerate : int
            The number of frames per second

        sim_step_min: int
            The number of physics steps per second. Increase this value if the physics is unstable and decrease it if the game runs slow.
        
        debug_draw: bool
            Whether or not to draw the collision shape for debugging purposes

        event_count: bool
            Whether or not to print out the number of active events for debugging purposes

        show_mouse_position: bool
            Whether or not to show the mouse position in the buttom-right corner

        exit_key: Optional[str]
            Very useful if you are working on a fullscreen game
            Set to None to disable it.

        saved_state_file: Optional[str]
            The path of the saved state. Default location will be used if set to None. 

        print_fps: bool
            Whether or not to print the fps

        use_frame_time: bool
            Use the number of frames to define game time. Note: highly experimental. 

        """

        if not (len(self.__screen_args) or len(self.__screen_kwargs)):
            self.__screen_kwargs = dict(size=(1280, 720))

        self._screen  = pygame.display.set_mode(*self.__screen_args, **self.__screen_kwargs)

        self._saved_states_manager.load_saved_state(saved_state_file)
        self.__framerate = framerate
        self.__screen_width = self._screen.get_width()
        self.__screen_height = self._screen.get_height()

        guide_lines_font = pygame.font.Font(None, 30)

        clock = pygame.time.Clock()

        

        draw_every_n_step = sim_step_min//framerate+1

        self._current_time_ms = 0

        if exit_key:
            self.when_key_pressed(exit_key).add_handler(lambda _: self.stop())
            
        self.create_pygame_event([pygame.QUIT]).add_handler(lambda _: self.stop())
        
        for t in self._game_start_triggers:
            t.trigger()

        cleanup_period = 2*framerate
        loop_count = 0

        threading.Thread(target=self._check_alive).start()
        frame_interval = 1000/framerate

        self.__start = True
        while self.__start:
            if print_fps:
                print(f"FPS: {clock.get_fps()}")            
            loop_count += 1
            
            dt = clock.tick(framerate)
            self._current_time_ms += frame_interval if use_frame_time else dt 
            for i in range(draw_every_n_step): 
                self._space.step(dt/draw_every_n_step)

            self._all_pygame_events = pygame.event.get()


            # check conditions
            for c in self._all_conditions:
                c._check()

            # execute 
            for t in self._all_triggers:
                t._handle_all(self._current_time_ms)
                # TODO: is it possible to remove t in the self.all_triggers here?
                t._generators_proceed(self._current_time_ms)

            # clean up
            self._all_conditions = list(filter(lambda t: t.stay_active, self._all_conditions))
            self._all_simple_key_triggers = list(filter(lambda t: t.stay_active, self._all_simple_key_triggers))
            self._all_triggers = list(filter(lambda t: t.stay_active, self._all_triggers))

            
            if event_count: 
                print("all_conditions", len(self._all_conditions))
                print("all_triggers", len(self._all_triggers))
                print("all sprite", len(self._all_sprites))
                # print("all_simple_key_triggers", len(self.all_simple_key_triggers))

            # Drawing

            #self._screen.fill((30, 30, 30))
            if not (self.__backdrop_key is None): 
                self._screen.blit(self.backdrops_by_key[self.__backdrop_key], (0, 0))
            else:
                self._screen.fill((255,255,255))
                helper._draw_guide_lines(self._screen, guide_lines_font, 100, 500)

                if show_mouse_position is None: 
                    helper._show_mouse_position(self._screen, guide_lines_font)

            if debug_draw: 
                self._space.debug_draw(self._draw_options)

            self._all_sprites.update()
            #self._all_sprites_to_show.update()
            self._all_sprites_to_show.draw(self._screen)

            if show_mouse_position:
                helper._show_mouse_position(self._screen, guide_lines_font)


            pygame.display.flip()
            if not loop_count % cleanup_period:
                self._do_autoremove()

    def _get_saved_state(self, sprite_id):
        return self._saved_states_manager.get_state_of(sprite_id)

    def save_sprite_states(self):
        """
        *EXTENDED FEATURE, EXPERIMENTAL*

        Save the x, y & direction of the sprites.

        To retrieve the states of the sprites, refer to [`retrieve_saved_state`](./sprite#Sprite.retrieve_saved_state)


        Usage:
        ```python
        from pyscratch import game

        # Start the game. The program stays in this line until the game window is closed. 
        game.start(60)

        # The program will get to this line when the game window is closed normally (not due to an error). 
        game.save_sprite_states()
        ```

        """
        self._saved_states_manager.save_sprite_states(self._all_sprites)

    def load_sound(self, key: str, path: str) :
        """
        Load the sound given a path, and index it with the key so it can be played later by `play_sound`

        Example:
        ```python
        game.load_sound('sound1', 'path/to/sound.wav')
        game.play_sound('sound1', volume=0.5)
        ```
        """
        if key in self._sounds: 
            raise KeyError(f'{key} already loaded. Choose a different key name.')
        
        self._sounds[key] = pygame.mixer.Sound(path)

    def play_sound(self, key:str, volume=1.0):
        """
        Play the sound given a key. 
        This method does not wait for the sound to finish playing. 

        Example:
        ```python
        game.load_sound('sound1', 'path/to/sound.wav')
        game.play_sound('sound1', volume=0.5)
        ```        
        """        
        s = self._sounds[key]
        s.set_volume(volume)
        s.play()
    
    
    def read_timer(self) -> float:
        """get the time (in seconds) since the game started."""
        return self._current_time_ms/1000
    

    def set_gravity(self, xy: Tuple[float, float]):
        """
        @private
        *EXTENDED FEATURE*
        
        Change the gravity of the space. Works for sprites with dynamic body type only, which is not the default.
        It will NOT work unless you explicitly make the sprite to have a dynamic body. 
        """
        self._space.gravity = xy

    def _new_sprite_of_file(self, caller_file):
        """
        return a index of the sprite created in the file
        """
        if not caller_file in self._sprite_count_per_file:
            self._sprite_count_per_file[caller_file] = 0
        else: 
            self._sprite_count_per_file[caller_file] += 1

        return self._sprite_count_per_file[caller_file]


    def _add_sprite(self, sprite: Sprite, caller_file=None):

        self._all_sprites.add(sprite)
        if len(self._all_sprites) > self.max_number_sprite:
            raise RuntimeError('Reached the maximum number sprite. ')
        #self._space.add(sprite.body, sprite.shape)
        self._sprite_click_trigger[sprite] = []
        # if to_show:
        #     sprite.show()
            #self._all_sprites_to_show.add(sprite)
        #sprite.update()

        if self.__started_interactive:
            sprite.set_draggable(True)

        return self._new_sprite_of_file(caller_file)

    def _cleanup_old_shape(self, old_shape):

        remove_list = []
        for pair in self._contact_pairs_set:
            if old_shape in pair:
                remove_list.append(pair)

        for r in remove_list:
            self._contact_pairs_set.remove(r)
        
    def _remove_sprite(self, sprite: Sprite):
        """
        Remove the sprite from the game.

        You can use the alias `Sprite.remove()` to do the same. 
        """

        self._all_sprites.remove(sprite)
        self._all_sprites_to_show.remove(sprite) 

        self._trigger_to_collision_pairs = {k: v for k, v in self._trigger_to_collision_pairs.items() if not sprite in v}

        
        self._cleanup_old_shape(sprite._shape)

        try: 
            self._space.remove(sprite._body, sprite._shape)
        except:
            print('removing non-existing shape or body')

        self._sprite_event_dependency_manager.sprite_removal(sprite)


    def _show_sprite(self, sprite: Sprite):
        """
        Show the sprite. 
        """
        self._all_sprites_to_show.add(sprite)

    def _hide_sprite(self, sprite: Sprite):
        """
        Hide the sprite. 
        """
        self._all_sprites_to_show.remove(sprite)

    def bring_to_front(self, sprite: Sprite):
        """
        Bring the sprite to the front. 
        Analogous to the "go to [front] layer" block in Scratch
        """
        #self._all_sprites_to_show.move_to_front(sprite)
        new_top_layer = self._all_sprites_to_show.get_top_layer()+1
        self.change_layer(sprite, new_top_layer)


    def move_to_back(self, sprite: Sprite):
        """
        Move the sprite to the back. 
        Analogous to the "go to [back] layer" block in Scratch
        """        
        self._all_sprites_to_show.move_to_back(sprite)
        sprite._assign_layer(0)

    def change_layer(self, sprite: Sprite, layer: int):
        """
        Bring the sprite to a specific layer. 
        """              
        self._all_sprites_to_show.change_layer(sprite, layer)
        sprite._assign_layer(layer)


    def change_layer_by(self, sprite: Sprite, by: int):
        """
        Analogous to the "go to [forward/backward] [N] layer" block in Scratch
        """           
        layer = self._all_sprites_to_show.get_layer_of_sprite(sprite)
        self._all_sprites_to_show.change_layer(sprite, layer + by)
        sprite._assign_layer(layer + by)


    def get_layer_of_sprite(self, sprite: Sprite):
        """
        Returns the layer number of the given sprite
        """
        self._all_sprites_to_show.get_layer_of_sprite(sprite)

    @deprecated("use add_backdrop")
    def set_backdrops(self, images: List[pygame.Surface]):
        """
        Set the list of all available backdrops. This function is meant to be run before the game start. 
        
        Example: 
        ```python
        # load the image into python 
        background_image = pysc.load_image('assets/my_background.jpg')
        background_image2 = pysc.load_image('assets/my_background2.jpg')
        background_image3 = pysc.load_image('assets/my_background3.jpg')

        # pass in a list of all the available backdrops. 
        pysc.game.set_backdrops([background_image, background_image2, background_image3])

        # choose the backdrop at index 1 (background_image2)
        pysc.game.switch_backdrop(1) 
        ```
        """
        for idx, img in enumerate(images):
            self.add_backdrop(str(idx), img)

    def add_backdrop(self, key, image: pygame.Surface):
        """
        Add the image as a backdrop, and index it with the key so it can be switched to using switch_backdrop 

        Example:
        ```python
        bg0 = pysc.load_image("assets/undersea_bg.png")'

        game.add_backdrop('background0', bg0)
        game.switch_backdrop('background0')
        ```

        """
        assert not key in self.backdrops_by_key, f"the name '{key}' is already in used"
        self.backdrop_keys.append(key)
        self.backdrops_by_key[key] = image
        self.backdrop_indices_by_key[key] = len(self.backdrop_keys) - 1 # minus one since the new one is added in the previous line
        

    @property
    def backdrop_index(self):
        """
        The index of the current backdrops. 
        """
        return self.__backdrop_index
    

    @property
    def backdrop_key(self):
        """
        The key (i.e. name) of the current backdrops.
        """
        return self.__backdrop_key
    
    def switch_backdrop(self, key:Optional[str]=None):
        """
        Change the backdrop by specifying the key of the backdrop.  
        """
        # backward compatibilty for v1
        if isinstance(key, int):
            key = str(key)
        if key != self.__backdrop_key:
            self.__backdrop_key = key
            self.__backdrop_index = None if key is None else self.backdrop_indices_by_key[key] 
            for t in self._backdrop_change_triggers:
                t.trigger(key)
            self._specific_backdrop_event_emitter.on_event(key)

    # TODO: testing needed
    def next_backdrop(self):
        """
        Switch to the next backdrop. 
        """
        if not self.__backdrop_index is None: 
            idx = (self.__backdrop_index+1) % len(self.backdrops_by_key)
            self.switch_backdrop(self.backdrop_keys[idx])
        
    # all events 

    ## scratch events

    def when_game_start(self, associated_sprites : Iterable[Sprite]=[])->Event[[]]:
        """
        It is recommended to use the `Sprite.when_game_start` alias instead of this method, 
        so you don't need to specify the `associated_sprites` in every event.

        Returns an `Event` that is triggered when you call `game.start`. 
        The event handler does not take in any parameter.

        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        

        t = self._create_event(associated_sprites)
        self._game_start_triggers.append(t)

        if TYPE_CHECKING:
            def sample_callback()-> Any:
                return
            t = _declare_callback_type(t, sample_callback)
        return t
            
    
    def when_any_key_pressed(self, associated_sprites : Iterable[Sprite]=[]) -> Event[[str, str]]:
        """
        It is recommended to use the `Sprite.when_any_key_pressed` alias instead of this method, 
        so you don't need to specify the `associated_sprites` in every event.  

        Returns an `Event` that is triggered when a key is pressed or released. 
        
        The event handler have to take two parameters:
        - **key** (str): The key that is pressed. For example, 'w', 'd', 'left', 'right', 'space'. 
            Uses [pygame.key.key_code](https://www.pygame.org/docs/ref/key.html#pygame.key.key_code) under the hood. 
        
        - **updown** (str): Either 'up' or 'down' that indicates whether it is a press or a release

        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
    
        """
        t = self._create_event(associated_sprites)
        self._all_simple_key_triggers.append(t)

        if TYPE_CHECKING:
            def sample_callback(key:str, updown:str)-> Any:
                return
            # this way the naming of the parameters is constrained too
            t = _declare_callback_type(t, sample_callback)
            #t = cast(Trigger[[str, str]], t)

        return t
    
    def when_key_pressed(self, key, associated_sprites : Iterable[Sprite]=[])-> Event[[str]]:
        """
        It is recommended to use the `Sprite.when_key_pressed` alias instead of this method, 
        so you don't need to specify the `associated_sprites` in every event.  
        
        Returns an `Event` that is triggered when a specific key is pressed or released. 

        The event handler have to take one parameter:
        - **updown** (str): Either 'up' or 'down' that indicates whether it is a press or a release
        
        Parameters
        ---
        key: str
            The key that triggers the event. For example, 'w', 'd', 'left', 'right', 'space'. 
            Uses [pygame.key.key_code](https://www.pygame.org/docs/ref/key.html#pygame.key.key_code) under the hood. 

        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        
        t = self._create_event(associated_sprites)

        if TYPE_CHECKING:
            def sample_callback(updown:str)-> Any:
                return
            # this way the naming of the parameters is constrained too
            t = _declare_callback_type(t, sample_callback)

        self._specific_key_event_emitter.add_event(key, t)
        return t    
    
    def when_this_sprite_clicked(self, sprite, other_associated_sprites: Iterable[Sprite]=[]) -> Event[[]]:
        """
        It is recommended to use the `Sprite.when_this_sprite_clicked` alias instead of this method.
        
        Returns an `Event` that is triggered when the given sprite is clicked by mouse.  
        The event handler does not take in any parameter.

        Parameters
        ---
        sprite: Sprite
            The sprite on which you want the click to be detected. The removal of this sprite will lead to the removal of this event so
            it does not need to be included in `other_assoicated_sprite`
        
        other_associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        
        t = self._create_event(set(list(other_associated_sprites)+[sprite]))

        if not sprite in self._sprite_click_trigger:
            self._sprite_click_trigger[sprite] = []
            
        self._sprite_click_trigger[sprite].append(t)
        if TYPE_CHECKING:
            def sample_callback()-> Any:
                return
            t = _declare_callback_type(t, sample_callback)
        return t
    

    def when_this_sprite_click_released(self, sprite, other_associated_sprites: Iterable[Sprite]=[]) -> Event[[]]:
        """
        The alias `Sprite.when_this_sprite_click_released` is not yet implemented. 

        Returns an `Event` that is triggered when the mouse click of the given sprite is released.  
        The event handler does not take in any parameter.

        Parameters
        ---
        sprite: Sprite
            The sprite on which you want the click to be detected. The removal of this sprite will lead to the removal of this event so
            it does not need to be included in `other_assoicated_sprite`
        
        other_associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        
        t = self._create_event(set(list(other_associated_sprites)+[sprite]))

        if not sprite in self._sprite_click_release_trigger:
            self._sprite_click_release_trigger[sprite] = []
            
        self._sprite_click_release_trigger[sprite].append(t)
        if TYPE_CHECKING:
            def sample_callback()-> Any:
                return
            t = _declare_callback_type(t, sample_callback)
        return t

    # TODO: testing needed
    def when_backdrop_switched(self, backdrop_key: str, associated_sprites : Iterable[Sprite]=[]) -> Event[[]]:
        """
        It is recommended to use the `Sprite.when_backdrop_switched` alias instead of this method, 
        so you don't need to specify the `associated_sprites` in every event.  

        Returns an `Event` that is triggered when the game is switched to a backdrop with the key `backdrop_key`.

        The event handler does not take in any parameter.

        Parameters
        ---
        backdrop_key: str
            The index of the backdrop  

        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        
                
        t = self._create_event(associated_sprites)

        if TYPE_CHECKING:
            def sample_callback()-> Any:
                return
            t = _declare_callback_type(t, sample_callback)

        self._specific_backdrop_event_emitter.add_event(backdrop_key, t)
        return t
    
    # TODO: testing needed
    def when_any_backdrop_switched(self, associated_sprites : Iterable[Sprite]=[]) -> Event[[Union[str,None]]]:
        """
        It is recommended to use the `Sprite.when_any_backdrop_switched` alias instead of this method, 
        so you don't need to specify the `associated_sprites` in every event.  
        
        Returns an `Event` that is triggered when the backdrop is switched. 

        The event handler have to take one parameter:
        - **str** (str): The key of the new backdrop  
        
        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        
        t = self._create_event(associated_sprites)
        self._backdrop_change_triggers.append(t)
        if TYPE_CHECKING:
            def sample_callback(key: Union[str, None])-> Any:
                return
            t = _declare_callback_type(t, sample_callback)

        return t

    def when_timer_above(self, t, associated_sprites : Iterable[Sprite]=[]) -> Condition:
        """
        It is recommended to use the `Sprite.when_timer_above` alias instead of this method, 
        so you don't need to specify the `associated_sprites` in every event.  
        
        Returns a `Condition` that is triggered after the game have started for `t` seconds.
        A `Condition` works the same way an `Event` does. 

        The event handler have to take one parameter:
        - **n** (int): This value will always be zero

        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        t = t*1000
        return self.when_condition_met(lambda:(self._current_time_ms>t), 1, associated_sprites)
    
    def when_started_as_clone(self, sprite, associated_sprites : Iterable[Sprite]=[]) -> Event[[Sprite]]:
        """
        Returns an `Event` that is triggered after the given sprite is cloned by `Sprite.create_clone`.
        Cloning of the clone will also trigger the event. 

        The event handler have to take one parameter:
        - **clone_sprite** (Sprite): The newly created clone.
                
        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        
        
        trigger = self._create_event(associated_sprites)
        self._clone_event_manager.new_trigger(sprite, trigger)
        if TYPE_CHECKING:
            def sample_callback(clone_sprite: Sprite)-> Any:
                return
            trigger = _declare_callback_type(trigger, sample_callback)
        return trigger

    def when_receive_message(self, topic: str, associated_sprites : Iterable[Sprite]=[]) -> Event[[Any]]:
        """
        It is recommended to use the `Sprite.when_receive_message` alias instead of this method, 
        so you don't need to specify the `associated_sprites` in every event.  

        Returns an `Event` that is triggered after a message of the given `topic` is broadcasted.

        The event handler have to take one parameter:
        - **data** (Any): This parameter can be anything passed on by the message.

        Parameters
        ---
        topic: str
            Can be any string. If the topic equals the topic of a broadcast, the event will be triggered. 
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        
        
        trigger = self._create_event(associated_sprites)
        self.__new_subscription(topic, trigger)
        if TYPE_CHECKING:
            def sample_callback(data: Any)-> Any:
                return
            trigger = _declare_callback_type(trigger, sample_callback)
        return trigger

    
    def broadcast_message(self, topic: str, data: Any=None):
        """
        Sends a message of a given `topic` and `data`.
        Triggers any event that subscribes to the topic. 
        The handlers of the events will receive `data` as the parameter.

        Example:
        ```python
        def event_handler(data):
            print(data) # data will be "hello world!"

        game.when_receive_message('print_message').add_handler(event_handler)
        game.broadcast_message('print_message', data='hello world!')

        # "hello world!" will be printed out
        ```
        Parameters
        ---
        topic: str
            Can be any string. If the topic of an message event equals the topic of the broadcast, the event will be triggered. 

        data: Any
            Any arbitory data that will be passed to the event handler
        
        """
        if not topic in self._all_message_subscriptions:
            return 
        
        self._all_message_subscriptions[topic] = list(filter(lambda t: t.stay_active, self._all_message_subscriptions[topic]))
        for e in self._all_message_subscriptions[topic]:
            e.trigger(data)

    def __new_subscription(self, topic: str, trigger: Event):
        if not (topic in self._all_message_subscriptions):
            self._all_message_subscriptions[topic] = []

        self._all_message_subscriptions[topic].append(trigger)

    def start_handler(self, handler:Optional[Callable[[], Any]]=None, associated_sprites: Iterable[Sprite]=[]):
        """
        *EXTENDED FEATURE*

        It is recommended to use the `Sprite.start_handler` alias instead of this method, 
        so you don't need to specify the `associated_sprites` in every event.  

        Run the event handler immediately. Useful when creating a sprite within a function.

        The handler does not take in any parameters. 

        Parameters
        ---
        handler: Function
            A function to run. 

        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        
        """
        e = self._create_event(associated_sprites)
        if handler: 
            e.add_handler(handler)
        e.trigger()
        return e


    ## advance events
    def create_pygame_event(self, types: List[int], associated_sprites : Iterable[Sprite]=[]) -> Event[[pygame.event.Event]]:
        """
        *EXTENDED FEATURE*

        Receives specific types of pygame events when they happen. 

        See pygame.event for more details: https://www.pygame.org/docs/ref/event.html

        ```python
        def key_press_event(event):
            # the event argument is a `pygame.event.Event`. 
            if event.type == pygame.KEYDOWN and event.key == pygame.K_d:
                print("the key d is down")

            if event.type == pygame.KEYUP and event.key == pygame.K_d:
                print("the key d is up")

        pysc.game.create_pygame_event([pygame.KEYDOWN, pygame.KEYUP])           
        ```

        The event handler have to take one parameter:
        - **event** (pygame.event.Event): An pygame event object. 

        Parameters
        ---
        types: List[int]
            A list of the pygame event flags. 

        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        
        
        condition = self.when_condition_met(associated_sprites)
        
        def checker_hijack():
            for e in self._all_pygame_events:
                if e.type in types:
                    condition.trigger.trigger(e)

            if not condition.trigger.stay_active:
                condition.remove()

        condition.change_checker(checker_hijack)

        return cast(Event[pygame.event.Event], condition.trigger)


    def _create_specific_collision_trigger(self, sprite1: Sprite, sprite2: Sprite, other_associated_sprites: Iterable[Sprite]=[]):
        """
        *EXTENDED FEATURE, EXPERIMENTAL*

        DOCUMENTATION NOT COMPLETED

        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        #"""Cannot change the collision type of the object after calling this function"""
        trigger = self._create_event(set(list(other_associated_sprites)+[sprite1, sprite2]))

        self._trigger_to_collision_pairs[trigger] = sprite1, sprite2

     
        return trigger

    def _create_type2type_collision_trigger(self, type_a:int, type_b:int, collision_suppressed=False, associated_sprites: Iterable[Sprite]=[]):
        """
        *EXTENDED FEATURE*

        DOCUMENTATION NOT COMPLETED

        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """        
        
        pair = (type_a, type_b) if type_a>type_b else (type_b, type_a)

        h = self._space.add_collision_handler(*pair)
        trigger = self._create_event(associated_sprites)



        if not pair in self._collision_type_pair_to_trigger: 
            self._collision_type_pair_to_trigger[pair] = []
        self._collision_type_pair_to_trigger[pair].append(trigger)

        collision_allowed = not collision_suppressed
        def begin(arbiter, space, data):
            game = cast(Game, data['game'])
            game._contact_pairs_set.add(arbiter.shapes) 

            for t in game._collision_type_pair_to_trigger[pair]:
                t.trigger(arbiter)
            return collision_allowed
        
        h.data['game'] = self
        h.begin = begin
        h.separate = _collision_separate

        
        return trigger


    def _create_type_collision_trigger(self, collision_type:int , collision_suppressed=False, associated_sprites: Iterable[Sprite]=[]):
        """
        *EXTENDED FEATURE*

        DOCUMENTATION NOT COMPLETED

        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        
        trigger = self._create_event(associated_sprites)

        collision_allowed = not collision_suppressed
        

        if not collision_type in self._collision_type_to_trigger: 
            self._collision_type_to_trigger[collision_type] = collision_allowed, []


        self._collision_type_to_trigger[collision_type][1].append(trigger)

        return trigger

    def _suppress_type_collision(self, collision_type, collision_suppressed=True):
        """
        *EXTENDED FEATURE*

        DOCUMENTATION NOT COMPLETED

        """
        collision_allowed = not collision_suppressed

        if not collision_type in self._collision_type_to_trigger: 
            self._collision_type_to_trigger[collision_type] = collision_allowed, []    
        else:
            t_list = self._collision_type_to_trigger[collision_type][1]
            self._collision_type_to_trigger[collision_type] = collision_allowed, t_list


    def when_timer_reset(self, reset_period: Optional[float]=None, repeats: Optional[int]=None, associated_sprites: Iterable[Sprite]=[]) -> TimerCondition:
        """
        *EXTENDED FEATURE*

        Repeats an event for `repeats` time for every `reset_period` seconds. 

        ```python
        
        def print_counts(n):
            print(n) # n is the number of remaining repeats

        # every one second, for 100 times
        pysc.game.when_timer_reset(1, 100).add_handler(print_counts)

        # will start printing 99, 98, ..., 0 every 1 second. 
        ```

        The event handler have to take one parameter:
        - **n** (int): The number of remaining repeats

        Parameters
        ---
        reset_period: float
            The reset period of the timer. The handlers are triggered on timer reset. 

        repeats: int or None
            How many times to repeat. Set to None for infinite repeats. 

        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """
        if reset_period is None: 
            reset_period = np.inf

        if repeats is None:
            _repeats = np.inf
        else:
            _repeats = repeats

        condition = TimerCondition(reset_period, _repeats)
        self._all_conditions.append(condition)
        self._all_triggers.append(condition.trigger)

        self._sprite_event_dependency_manager.add_event(
            condition, associated_sprites
        )                
        return condition
    
    
    def when_condition_met(self, checker=lambda: False,  repeats: Optional[int]=None,  associated_sprites: Iterable[Sprite]=[])-> Condition:
        """
        *EXTENDED FEATURE*

        For every frame, if a condition is met, the event is triggered. Repeated up to `repeats` times. 

        The condition is provided by a function that takes no argument and returns a boolean. 
        
        ```python
        def slowly_move_sprite_out_of_edge(n):
            my_sprite.x += 1
            
        pysc.game.when_condition_met(lambda: (my_sprite.x<0), None).add_handler(slowly_move_sprite_out_of_edge)
        ```

        The event handler have to take one parameter:
        - **n** (int): The number of remaining repeats

        Parameters
        ---
        checker: Callable[[], bool] 
            A function that takes no argument and returns a boolean. 
            The checker is run one every frame. If it returns true, the handler is called. 

        repeats: int or None
            How many times to repeat. Set to None for infinite repeats. 

                    
        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """

        if repeats is None:
            _repeats = np.inf
        else:
            _repeats = repeats

        condition = Condition(checker, _repeats)
        self._all_conditions.append(condition)
        self._all_triggers.append(condition.trigger)

        self._sprite_event_dependency_manager.add_event(
            condition, associated_sprites
        )        
        return condition
    
    def when_mouse_click(self, associated_sprites: Iterable[Sprite]=[] ) -> Event[[Tuple[int, int], int, str]]:
        """
        *EXTENDED FEATURE*

        Returns an `Event` that is triggered when the mouse is clicked or released. 
        
        The event handler have to take three parameters:
        - **pos** (Tuple[int, int]): The location of the click
        - **button** (int): Indicates which button is clicked. 0 for left, 1 for middle, 2 for right and other numbers for scrolling.     
        - **updown** (str): Either 'up' or 'down' that indicates whether it is a press or a release

        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
    
        """
        event_internal = self.create_pygame_event([pygame.MOUSEBUTTONUP, pygame.MOUSEBUTTONDOWN], associated_sprites)
        event = self._create_event(associated_sprites)
        def handler(e: pygame.event.Event):
            updown = "up" if e.type == pygame.MOUSEBUTTONUP else "down"
            event.trigger(e.pos, e.button, updown)
        event_internal.add_handler(handler)

        return event
    
    def when_mouse_scroll(self, associated_sprites: Iterable[Sprite]=[] ) -> Event[[str]]:
        """
        *EXTENDED FEATURE*

        Returns an `Event` that is triggered when the mouse is scrolled. 
        
        The event handler have to take one parameters:
        - **updown** (str): Either 'up' or 'down' that indicates whether it is a scroll up or a scroll down.

        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
    
        """
        event_internal = self.create_pygame_event([pygame.MOUSEWHEEL], associated_sprites)
        event = self._create_event(associated_sprites)
        def handler(e):
            updown = "up" if e.y > 0 else "down"
            event.trigger(updown)
        event_internal.add_handler(handler)

        return event
    

    def _create_event(self, associated_sprites: Iterable[Sprite]=[]) -> Event:
        """
        
        Parameters
        ---
        associated_sprites: List[Sprite]
            A list of sprites that this event depends on. Removal of any of these sprites leads to the removal of the event. 
        """

        trigger = Event()
        self._all_triggers.append(trigger)

        self._sprite_event_dependency_manager.add_event(
            trigger, associated_sprites
        )
        return trigger
    

game = Game()
"""
The singleton Game object. This is the object that represent the game.   
"""


def is_key_pressed(key: str) -> bool:
    """
    Returns a bool(True/False) that indicates if the given key is pressed.

    Usage
    ```python
    if is_key_pressed('space'):
        print('space pressed')
    ```
    """
    keycode = pygame.key.key_code(key)
    result = pygame.key.get_pressed()
    return result[keycode]


def get_mouse_pos() -> Tuple[int, int]:
    """
    Returns the mouse coordinate. 
    
    Usage
    ```python
    mouse_x, mouse_y = get_mouse_pos()
    ```
    """
    return pygame.mouse.get_pos()


def get_mouse_presses() -> Tuple[bool, bool, bool]:
    """
    Returns the mouse presses. 
    ```python
    is_left_click, is_middle_click, is_right_click = get_mouse_presses()
    ```
    """
    return pygame.mouse.get_pressed(num_buttons=3)

# def _is_touching(sprite_a:Sprite, sprite_b:Sprite):
#     """
#     pymunk
#     """
#     for pair in game._contact_pairs_set:

#         if (sprite_a._shape in pair) and (sprite_b._shape in pair):
#             return True
#     return False


# def _is_touching_mouse(sprite: Sprite):
#     return sprite._shape.point_query(pygame.mouse.get_pos()).distance <= 0
        


"""
pysc.game.broadcast_message
    ("cut_sprite_frame_drop", dict(sprite=sprite, position=pysc.get_mouse_pos()))

"""
from pathlib import Path
from typing import List, Tuple, cast

import numpy as np
from pygame import Surface
import pyscratch as pysc
from settings import *
game = pysc.game
width, height = RIGHT_PANEL_WIDTH, SCREEN_HEIGHT - PANEL_MARGIN - BOTTOM_RIGHT_PANEL_HEIGHT

colour = 255, 255, 255, 255
ss_view = pysc.create_rect_sprite(colour, width, height)

ss_view.set_xy((SCREEN_WIDTH - PANEL_MARGIN - RIGHT_PANEL_WIDTH/2,  SCREEN_HEIGHT/2-BOTTOM_RIGHT_PANEL_HEIGHT/2))
#ss_view.set_draggable(False)
#topleft = ss_view.x-width/2, ss_view.y-height/2
pysc.game['ss_view'] = ss_view

def topleft():
    return ss_view.x-width/2, ss_view.y-height/2
ss_view['frame_list'] = []

pysc.game['ss_view_topleft'] = topleft()
pysc.game['ss_view_buttom_right'] = ss_view.x+width/2, ss_view.y+height/2



ss_view['spritesheet_sprite'] = None
pysc.game['ss_sprite'] = None
def on_msg_mode_change(mode):
    if not mode == 'nav':
        ss_view.show()
        pysc.game.move_to_back(ss_view)
        for f in ss_view.sprite_data['frame_list']:
            f.show()

    else:
        ss_view.hide()
        for f in ss_view.sprite_data['frame_list']:
            f.hide()
        if ss_sprite := ss_view['spritesheet_sprite']:
            ss_sprite.hide()
            

ss_view.when_receive_message('cut_or_nav_mode_change').add_handler(on_msg_mode_change)

def try_load_image(path):
    try: 
        return pysc.load_image(path)
    except:
        return None


def on_msg_image_selected(path):
    
    for f in ss_view['frame_list']:
        f.remove()
    ss_view['frame_list'] = []

    img = try_load_image(path)
    if img: 
        pysc.game.shared_data['image_on_right_display'] = img

        if ss_sprite := ss_view['spritesheet_sprite']:
            ss_sprite.remove()

        ss_view['spritesheet_sprite'] = pysc.Sprite({'a':[img]})

        ss_sprite:pysc.Sprite = ss_view['spritesheet_sprite']

        pysc.game['ss_sprite'] = ss_sprite
        ss_sprite.oob_limit=np.inf
        ss_sprite['original_width'] = img.get_width()
        ss_sprite['original_height'] = img.get_height()
        game.change_layer(ss_sprite, 0)

        
        #ss_sprite.lock_to(ss_view, offset=(0,0)) 
        #ss_sprite.set_xy((0,0))

        ss_sprite.set_xy((ss_view.rect.centerx, ss_view.rect.centery))
        ss_sprite.set_draggable(True) #TODO: dragging a locked sprite leads to unexpected behaviour

        ss_sprite._drawing_manager.set_mask_threshold(-1)


        pysc.game.broadcast_message('cut_or_nav_mode_change', 'cut')
    pass

ss_view.when_receive_message('image_selected').add_handler(on_msg_image_selected)
from itertools import product
def on_msg_cut(_):
    ss_view.draw(pysc.create_rect((0,0,0,0), 1, 1)) # reset

    if not 'image_on_right_display' in pysc.game.shared_data:
        pysc.game.broadcast_message('warning', 'image not selected' )
        return 
    
    for f in ss_view['frame_list']:
        f.remove()
    ss_view['frame_list'] = []

    
    if ss_sprite := ss_view['spritesheet_sprite']:
        ss_sprite.hide()
    

    spritesheet: pygame.Surface = pysc.game.shared_data['image_on_right_display'] 

    n_row =  pysc.game.shared_data['n_row'] 
    n_col =  pysc.game.shared_data['n_col'] 

    if not n_row:
        print('invalid n_row')
        pysc.game.broadcast_message('warning', 'invalid n_row' )
        return
    
    if not n_col:
        print('invalid n_col')
        pysc.game.broadcast_message('warning', 'invalid n_col' )
        return
    n_row = int(n_row)
    n_col = int(n_col)
    print(game['offset_x'], game['offset_y'], game['size_x'], game['size_y'])
    spritesheet_to_cut = spritesheet.subsurface(game['offset_x'], game['offset_y'], game['size_x'], game['size_y'])

    for i, (r, c) in enumerate(product(range(n_row), range(n_col))):
        frame = pysc.get_frame_from_sprite_sheet_by_frame_size(spritesheet_to_cut, game['pixel_x'], game['pixel_y'], c, r)
        sprite = SpriteFrameAfterCut(frame, i, ss_sprite.scale_factor, n_col)
        ss_view['frame_list'].append(sprite)

ss_view.when_receive_message('cut').add_handler(on_msg_cut)



def SpriteFrameAfterCut(surface: Surface, order, scale_factor, n_col):
    w, h = surface.get_width()*scale_factor, surface.get_height()*scale_factor
    sprite = pysc.Sprite({'always':[surface]})
    sprite.oob_limit = np.inf
    sprite.scale_by(scale_factor)


    spacing = 0
    lt = topleft()
    x = spacing+(order%n_col)*(w + spacing)+lt[0] +w/2
    y = spacing+(order//n_col)*(h+spacing)+lt[1] +h/2
    
    sprite['x'] = x
    sprite['y'] = y

    sprite.set_xy((x,y))
    sprite.set_draggable(True) 



    def on_mouse_release():

        pysc.game.broadcast_message("cut_sprite_frame_drop", dict(surface=surface, sprite=sprite, position=pysc.get_mouse_pos()))
        pass


    
    pysc.game.when_this_sprite_click_released(sprite).add_handler(on_mouse_release)


    return sprite



# TODO: The scrolling becomes hard to use because touching only happens in the non-transparent pixels


def on_scroll(updown):
    #if updown.
    ss_sprite:pysc.Sprite = ss_view['spritesheet_sprite']
    if not ss_sprite: return
    if not ss_sprite.is_touching_mouse():
        return
    
    if updown == 'up':
        ss_sprite.scale_by(1.05)
    else:
        ss_sprite.scale_by(1/1.05)
    
    pysc.game.broadcast_message("ss_sprite_scale_change")

pysc.game.when_mouse_scroll([ss_view]).add_handler(on_scroll)



def draw_cutting_rect(surface, colour, x0, y0, x1, y1, n_row, n_col):

    n_row = 1 if not n_row else n_row
    n_col = 1 if not n_col else n_col

    x0, x1 = min(x0, x1), max(x0, x1)
    y0, y1 = min(y0, y1), max(y0, y1)

    n_line_h = n_row - 1
    n_line_v = n_col - 1


    pygame.draw.lines(surface, colour, True, [(x0, y0), (x0, y1), (x1, y1), (x1, y0)])

    x_step = (x1 - x0)/n_col # vertical lines
    y_step = (y1 - y0)/n_row # horizontal lines
    # 
    for lx in range(int(n_line_v)):
        xpos = (1+lx)*x_step 
        xpos += x0
        pygame.draw.line(surface, colour, (xpos, y0), (xpos, y1))

    for ly in range(int(n_line_h)):
        ypos = (1+ly)*y_step 
        ypos += y0
        pygame.draw.line(surface, colour, (x0, ypos), (x1, ypos))

    return x_step, y_step



    
def draw_selection_rect():
    cir0, cir1 = pysc.game['c0'], pysc.game['c1']
    while True: 
        yield 1/30
        ss_sprite:pysc.Sprite = ss_view['spritesheet_sprite']
        if not ss_sprite: 
            continue

        if game['pixel_x'] is None: continue
        if game['pixel_y'] is None: continue

        ss_view._drawing_manager.frames[0].fill((255,255,255))

        n_col = 1 if not pysc.game['n_col'] else pysc.game['n_col']
        n_row = 1 if not pysc.game['n_row'] else pysc.game['n_row']


        # top left of the selected rect
        gx0 =  game['offset_x']*ss_sprite.scale_factor + ss_sprite.rect.left  - ss_view.rect.left
        gy0 = game['offset_y']*ss_sprite.scale_factor + ss_sprite.rect.top  - ss_view.rect.top
        
        gx1 = gx0 + game['pixel_x']*n_col*ss_sprite.scale_factor 
        gy1 = gy0 + game['pixel_y']*n_row*ss_sprite.scale_factor


        #pygame.draw.lines(ss_view._drawing_manager.frames[0], (0,0,0), True, [(gx0, gy0), (gx0, gy1), (gx1, gy1), (gx1, gy0)])

        draw_cutting_rect(
            ss_view._drawing_manager.frames[0],
            (0,0,0),
            gx0, gy0, gx1,  gy1,
            pysc.game['n_row'], pysc.game['n_col']
        )       

ss_view.when_game_start().add_handler(draw_selection_rect)



import pyscratch as pysc

# create the sprites
heart1 = pysc.create_animated_sprite("assets/hearts")
heart2 = pysc.create_animated_sprite("assets/hearts")
heart3 = pysc.create_animated_sprite("assets/hearts")

# variable definitons should be done outside the event to guarantee the variables is defined before any event try to access it 
pysc.game['health'] = 3

def heart_display():

    heart1.set_xy((100, 100))
    heart1.set_scale(3)

    heart2.set_xy((155, 100))
    heart2.set_scale(3)

    heart3.set_xy((210, 100))
    heart3.set_scale(3)


    while True: 
        if pysc.game['health'] < 3:
            heart3.set_frame(1)

        if pysc.game['health'] < 2:
            heart2.set_frame(1)

        if pysc.game['health'] < 1:
            heart1.set_frame(1)
        
        yield 1/pysc.game['framerate']
    
# ideally, you should do
# `pysc.game.when_game_start([heart1, heart2, heart3])`, so when any of the hearts is removed, the event is removed as well. 
# but we aren't going to remove any of the hearts, so there's no need to do so there. 

pysc.game.when_game_start().add_handler(heart_display)



import pyscratch as pysc
from pyscratch import game
import player, platforms

width = 720
height = 1280

game.update_screen_mode((width, height))
game.start(60, show_mouse_position=True, print_fps=False)


from calendar import Day
import pyscratch as pysc
from settings import *

w, h = 120, 40
selected_colour = (255, 255, 255)
deselected_colour = (127, 127, 127)


def FloatInputBoxVerticallyLabel(data_key, message_on_change="", default_value=""):
    label = pysc.create_rect_sprite(deselected_colour, w, h)
    label.write_text(data_key, DEFAULT_FONT24, offset=(w/2, h/2))

    selected_sur = pysc.create_rect(selected_colour, w, h)
    deselected_sur = pysc.create_rect(deselected_colour, w, h)
    pysc.game.shared_data[data_key] = None
    text_box = pysc.Sprite({'selected':[selected_sur], 'deselected':[deselected_sur]}, 'deselected')
    label.lock_to(text_box, (0,0), reset_xy=True)
    label.y = -h

    # def on_click():
    #     text_box.set_animation('selected')
    #     text_box.sprite_data['selected'] = True

    # text_box.when_this_sprite_clicked().add_handler(on_click)
    text_box['selected'] = False
    text_box['text'] = default_value
    try:
        pysc.game.shared_data[data_key] = float(text_box.sprite_data['text'])
    except:
        pysc.game.shared_data[data_key] = None    

    text_box.write_text(text_box['text'], DEFAULT_FONT24, colour=(255,255,255), offset=(w/2, h/2))
    def on_any_key_press(key:str, updown):
        if updown == 'up': return
        if not text_box.sprite_data['selected']: return
        if text_box['just_selected']: 
            text_box['just_selected'] = False
            text_box['text'] = ''

        if key.isdigit() or key == '.':
            text_box.sprite_data['text'] += key

        if key == 'backspace' and len(text_box.sprite_data['text']):
            text_box.sprite_data['text'] = text_box.sprite_data['text'][:-1]

        try:
            pysc.game.shared_data[data_key] = float(text_box.sprite_data['text'])
        except:
            pysc.game.shared_data[data_key] = None
        
        text_box.write_text(text_box.sprite_data['text'], DEFAULT_FONT24, colour=(0,0,0), offset=(w/2, h/2))

        if message_on_change:
            pysc.game.broadcast_message(message_on_change, pysc.game.shared_data[data_key])
            
    text_box.when_any_key_pressed().add_handler(on_any_key_press)

    def on_any_mouse_click(pos, button, updown):
        if not text_box.is_touching_mouse():
            #print(button)
            #print(pos)
            text_box.sprite_data['selected'] = False
            text_box['just_selected'] = False
            text_box.set_animation('deselected')
            text_box.write_text(text_box.sprite_data['text'], DEFAULT_FONT24, colour=(255,255,255), offset=(w/2, h/2))

        else:
            text_box.sprite_data['selected'] = True
            text_box['just_selected'] = True
            text_box.set_animation('selected')
            text_box.write_text(text_box['text'], DEFAULT_FONT24, colour=(0,0,0), offset=(w/2, h/2))
    pysc.game.when_mouse_click([text_box]).add_handler(on_any_mouse_click)

    return text_box


def FloatInputBoxHoriLabel(data_key, message_on_change="", default_value="", label_width=80):
    wl = label_width
    label = pysc.create_rect_sprite(deselected_colour, wl, h)
    label.write_text(data_key, DEFAULT_FONT24, offset=(wl/2, h/2))

    selected_sur = pysc.create_rect(selected_colour, w, h)
    deselected_sur = pysc.create_rect(deselected_colour, w, h)
    pysc.game.shared_data[data_key] = None
    text_box = pysc.Sprite({'selected':[selected_sur], 'deselected':[deselected_sur]}, 'deselected')
    label.lock_to(text_box, (0,0), reset_xy=True)
    label.x = -w/2-wl/2


    # def on_click():
    #     text_box.set_animation('selected')
    #     text_box.sprite_data['selected'] = True

    # text_box.when_this_sprite_clicked().add_handler(on_click)
    text_box['selected'] = False
    text_box['text'] = default_value
    try:
        pysc.game.shared_data[data_key] = float(text_box.sprite_data['text'])
    except:
        pysc.game.shared_data[data_key] = None    

    text_box.write_text(text_box['text'], DEFAULT_FONT24, colour=(255,255,255), offset=(w/2, h/2))
    def on_any_key_press(key:str, updown):
        if updown == 'up': return
        if not text_box.sprite_data['selected']: return
        if text_box['just_selected']: 
            text_box['just_selected'] = False
            text_box['text'] = ''

        if key.isdigit() or key == '.':
            text_box.sprite_data['text'] += key

        if key == 'backspace' and len(text_box.sprite_data['text']):
            text_box.sprite_data['text'] = text_box.sprite_data['text'][:-1]

        try:
            pysc.game.shared_data[data_key] = float(text_box.sprite_data['text'])
        except:
            pysc.game.shared_data[data_key] = None
        
        text_box.write_text(text_box.sprite_data['text'], DEFAULT_FONT24, colour=(0,0,0), offset=(w/2, h/2))

        if message_on_change:
            pysc.game.broadcast_message(message_on_change, pysc.game.shared_data[data_key])
            
    text_box.when_any_key_pressed().add_handler(on_any_key_press)

    def on_any_mouse_click(pos, button, updown):
        if not text_box.is_touching_mouse():
            #print(button)
            #print(pos)
            text_box.sprite_data['selected'] = False
            text_box['just_selected'] = False
            text_box.set_animation('deselected')
            text_box.write_text(text_box.sprite_data['text'], DEFAULT_FONT24, colour=(255,255,255), offset=(w/2, h/2))

        else:
            text_box.sprite_data['selected'] = True
            text_box['just_selected'] = True
            text_box.set_animation('selected')
            text_box.write_text(text_box['text'], DEFAULT_FONT24, colour=(0,0,0), offset=(w/2, h/2))
    pysc.game.when_mouse_click([text_box]).add_handler(on_any_mouse_click)

    return text_box


def IntegerInputBox(data_key, message_on_change="", default_value="", label_width=60):
    wl = label_width
    label = pysc.create_rect_sprite(deselected_colour, wl, h)
    label.write_text(data_key, DEFAULT_FONT24, offset=(wl/2, h/2))

    selected_sur = pysc.create_rect(selected_colour, w, h)
    deselected_sur = pysc.create_rect(deselected_colour, w, h)
    pysc.game.shared_data[data_key] = None
    text_box = pysc.Sprite({'selected':[selected_sur], 'deselected':[deselected_sur]}, 'deselected')
    label.lock_to(text_box, (0,0), reset_xy=True)
    label.x = -w/2-wl/2

    # def on_click():
    #     text_box.set_animation('selected')
    #     text_box.sprite_data['selected'] = True

    # text_box.when_this_sprite_clicked().add_handler(on_click)
    text_box.sprite_data['selected'] = False
    text_box.sprite_data['text'] = default_value
    text_box.write_text(text_box['text'], DEFAULT_FONT24, colour=(255,255,255), offset=(w/2, h/2))
    def on_any_key_press(key:str, updown):
        if updown == 'up': return
        if not text_box.sprite_data['selected']: return
        if text_box['just_selected']: 
            text_box['just_selected'] = False
            text_box['text'] = ''

        if key.isdigit():
            text_box.sprite_data['text'] += key

        if key == 'backspace' and len(text_box.sprite_data['text']):
            text_box.sprite_data['text'] = text_box.sprite_data['text'][:-1]

        # try:
        #     pysc.game.shared_data[data_key] = int(text_box.sprite_data['text'])
        # except:
        #     pysc.game.shared_data[data_key] = None
        
        text_box.write_text(text_box.sprite_data['text'], DEFAULT_FONT24, colour=(0,0,0), offset=(w/2, h/2))

        
        
    text_box.when_any_key_pressed().add_handler(on_any_key_press)

    def on_any_mouse_click(pos, button, updown):
        if not text_box.is_touching_mouse():
            #print(button)
            #print(pos)
            if text_box['selected']:
                try:
                    pysc.game.shared_data[data_key] = int(text_box.sprite_data['text'])
                except:
                    pysc.game.shared_data[data_key] = None
                if message_on_change:
                    pysc.game.broadcast_message(message_on_change, pysc.game.shared_data[data_key])
            text_box['selected'] = False
            text_box['just_selected'] = False
            text_box.set_animation('deselected')
            text_box.write_text(text_box['text'], DEFAULT_FONT24, colour=(255,255,255), offset=(w/2, h/2))

        else:
            text_box['selected'] = True
            text_box['just_selected'] = True
            text_box.set_animation('selected')
            text_box.write_text(text_box['text'], DEFAULT_FONT24, colour=(0,0,0), offset=(w/2, h/2))
    pysc.game.when_mouse_click([text_box]).add_handler(on_any_mouse_click)


    def update_value_on_change():
        while True:
            yield 1/30

            

            if text_box.animation_name == "selected":
                continue

            try: 
                text_box['text'] = str(round(pysc.game[data_key]))
            except:
                text_box['text'] = ""

            text_box.write_text(text_box['text'], DEFAULT_FONT24, colour=(255,255,255), offset=(w/2, h/2))


    text_box.when_game_start().add_handler(update_value_on_change)

    return text_box


"""
Everything in this module is directly under the pyscratch namespace. 
For example, instead of doing `pysc.event.Timer`,
you can also directly do `pysc.Timer`.
"""

from __future__ import annotations
from typing import Any, Callable, Dict, Generator, Generic, List, ParamSpec, Set, Tuple, TypeVar, Union, override
from types import GeneratorType, NoneType

import pygame
import pymunk
import numpy as np

P = ParamSpec('P')
"""@private"""
T = TypeVar('T')
"""@private"""


def _declare_callback_type(obj: Event, func: Callable[P, Union[NoneType, Generator[float, None, None]]]) -> Event[P]:
    return obj


class Event(Generic[P]):
    """
    A low level Event class. 
    You **do not** need to create the Event object yourself. 
    They are instead returned by functions like `sprite.when_game_started()`

    In no circumstances will you need to create your custom event. Always use the message event instead. 
    """

    def __init__(self):
        """@private"""
        
        self.__triggers: List = []
        self.__callbacks: List[Callable[P, Union[NoneType, Generator[float, None, None]]]] = []
        self.__stay_active = True
        self.__to_soft_remove = False
        self.__generators: Dict[GeneratorType, float] = {}
        

    #def add_handler(self, func: Callable[P, Union[NoneType, Generator[float, None, None]]]): 
    # use Any for the simplicity in the doc
    def add_handler(self, func: Callable[P, Any]):
        """
        Add a handler function to this event. 
        When the event is triggered, the handler functions are called, taking in the parameters passed on by the triggers.
        """
        self.__callbacks.append(func)
        return self

    def __call__(self, func: Callable[P, Any]):
        return self.add_handler(func)
            
    


    def remove(self):
        """Schedule this event to be removed."""
        self.__stay_active = False

    def soft_remove(self):
        """Schedule this event to be removed when all the generators (yield functions) finish"""
        self.__to_soft_remove = True

    @property
    def stay_active(self) -> bool:
        """Shows whether this event is removed or scheduled to be removed."""
        return self.__stay_active
    


    def trigger(self, *args: P.args, **kwargs: P.kwargs):
        """
        @private
        Trigger the event. The parameters will be passed on to the event handlers.

        You will not need to run this method ever.
        """
        self.__triggers.append((args, kwargs))
    
    def _handle_all(self, current_t_ms):
        while self._handle_one(current_t_ms):
            pass
        
    def _handle_one(self, current_t_ms):

        if not len(self.__triggers): 
            return False
        
        args, kwargs = self.__triggers.pop(0)

        for cb in self.__callbacks:
            ret = cb(*args, **kwargs) # type: ignore
            if isinstance(ret, GeneratorType):

                # just to schedule to run immediately 
                # actually run when self.generators_proceed is called
                self.__generators[ret] = current_t_ms

        return True
    
    def _generators_proceed(self, current_t_ms):
        to_remove = []

        for g, t in self.__generators.items():
            next_t = t
            while next_t<=current_t_ms:
                try:
                    next_t = next_t+next(g)*1000
                    self.__generators[g] = next_t
                except StopIteration:
                    next_t = np.inf
                    to_remove.append(g)

        self.__generators = {g:t for g, t in self.__generators.items() if not g in to_remove}
        if not len(self.__generators) and self.__to_soft_remove:
            self.remove()



class _ConditionInterface:
    """
    A low level Condition class. 
    You **do not** need to create the Condition object yourself. 
    """
    def _check(self):
        pass

    def add_handler(self, callback: Callable[[int], Any]):
        """
        Add a handler function to this event. 
        When the event is triggered, the handler functions are called, taking in the parameters passed on by the triggers.
        """        
        return self
        
    def remove(self):
        """Schedule this event to be removed."""
        pass

    def soft_remove(self):
        """Schedule this event to be removed when all the generators (yield functions) finish"""
        pass

    @property
    def stay_active(self) -> bool:
        """Shows whether this event is removed or scheduled to be removed."""
        return True

class Condition(_ConditionInterface):
    def __init__(self, checker: Callable[[], bool] = lambda: False, repeats: Union[float, int]=1):
        """@private"""
        self.trigger = Event()
        self.repeat_remains = repeats
        self.checker = checker
        self.__stay_active = True

    def add_handler(self, callback: Callable[[int], Any]):       
        self.trigger.add_handler(callback)
        return self
    
    def __call__(self, callback: Callable[[int], Any]):
        return self.add_handler(callback)
                
    def remove(self):
        self.__stay_active = False
        self.trigger.remove()

    def soft_remove(self):
        self.__stay_active = False
        self.trigger.soft_remove()

    @property
    def stay_active(self):
        return self.__stay_active

    def _check(self):
        if self.checker() and self.repeat_remains:
            self.repeat_remains -= 1
            self.trigger.trigger(self.repeat_remains)

        if not self.repeat_remains:
            self.soft_remove()

    def change_checker(self, checker= lambda: False):
        """@private"""
        self.checker = checker

    

class TimerCondition(_ConditionInterface):
    def __init__(self, reset_period=np.inf, repeats=np.inf):
        """
        @private
        """
        self.trigger = Event()
        self.repeat_remains = repeats
        self.timer = Timer(reset_period=reset_period)
        self.period = 0
        self.__stay_active = True

    def add_handler(self, callback: Callable[[int], Any]):
        self.trigger.add_handler(callback)
        return self
    
    def __call__(self, callback: Callable[[int], Any]):
        return self.add_handler(callback)
                
    def remove(self):
        self.__stay_active = False
        self.trigger.remove()

    def soft_remove(self):
        self.__stay_active = False
        self.trigger.soft_remove()

    @property
    def stay_active(self):
        return self.__stay_active

    def _check(self):
        self.timer.read()
        while (self.timer.n_period > self.period) and self.repeat_remains:
            self.period += 1
            self.repeat_remains -= 1
            self.trigger.trigger(self.repeat_remains)

        if not self.repeat_remains:
            self.soft_remove()
        


class Timer:
    """
    Create a timer that resets every `reset_period` seconds. 
    """
    def __init__(self, reset_period=np.inf):
        """
        Parameters
        ---
        reset_period: float
            The reset period. Pass in np.inf for no reset
        """
        self._t0 = pygame.time.get_ticks()/1000
        self.reset_period = reset_period
        self.n_period = 0
        """How many times the `reset_period` have passed"""

    def read(self):
        """
        Read the timer.
        """
        dt = pygame.time.get_ticks()/1000 - self._t0
        self.n_period = int(dt // self.reset_period)
        return dt % self.reset_period
    
    def full_reset(self):
        """
        Reset the timer and the `n_period`.
        """
        self.n_period = 0
        self._t0 = pygame.time.get_ticks()/1000

    

        

        






import pyscratch as pysc
from pyscratch import game

chest = pysc.create_single_costume_sprite("assets/chest-open.png")

@chest.when_this_sprite_clicked()
def my_click_event():
    message = "I am a chest! "
    n_repeat = 4
    repeated_message = message * n_repeat
    print(repeated_message)



"""
msg in:
    show_bin
    hide_bin

shared data
    frame_bin
"""


import pyscratch as pysc
from settings import *


frame_bin = pysc.create_rect_sprite((255, 127, 127), 100, 100)
frame_bin.write_text("bin", DEFAULT_FONT48, offset=(50, 50))

def set_xy():
    
    container = pysc.game['main_bottom_panel']

    frame_bin.hide()
    frame_bin.lock_to(container, (0,0))
    frame_bin.x = 0
    frame_bin.y = 0
    
frame_bin.when_game_start().add_handler(set_xy)


frame_bin.when_receive_message('show_bin').add_handler(lambda d: frame_bin.show())
frame_bin.when_receive_message('hide_bin').add_handler(lambda d: frame_bin.hide())

pysc.game['frame_bin'] = frame_bin


import pyscratch as pysc
from pyscratch import game
import chest, enemy, friend
       
# background
bg0 = pysc.load_image("assets/undersea_bg.png")
bg0 = pysc.scale_to_fit_aspect(bg0, (1024, 576))
game.add_backdrop('sea', bg0)

bg1 = pysc.load_image("assets/Cat In Space Wallpaper Hq.jpg")
bg1 = pysc.scale_to_fit_aspect(bg1, (1024, 576))
game.add_backdrop('lose', bg1)


game.switch_backdrop('sea')



def background():
    while True:
        yield 1/game.framerate
        if game['score'] < 0: 
            game.switch_backdrop('lose')
game.when_game_start().add_handler(background)


# starting the game
game.update_screen_mode((1024, 576))
game.start(show_mouse_position=True)





import pyscratch as pysc
from pyscratch import game
#pysc.create_shared_data_display_sprite("speed_multiplier")

perk = pysc.create_single_costume_sprite("assets/tool_bomb.png")

def when_perk_clicked():
    """
    reduce speed multiplier when perk clicked
    """
    perk.hide()
    perk.x = pysc.random_number(0, game.screen_width/2)
    perk.y = pysc.random_number(0, game.screen_height/2)
    game['speed_multiplier'] *= 0.9

perk.when_this_sprite_clicked().add_handler(when_perk_clicked)




def game_start():
    """
    perk appears
    """
    while True:
        yield pysc.random_number(1, 3)
        perk.show()

perk.when_game_start().add_handler(game_start)





"""
broadcast messages: 
    scrolling (scrolling_offset)

shared_data:
    scrolling_offset
"""


from pathlib import Path
from typing import Dict

from pygame import Surface
import pyscratch as pysc
from settings import *



# Sprite: the panel (the background of the preview frames)
PANEL_BG_COLOUR = 199, 207, 224
panel = pysc.create_rect_sprite(
    PANEL_BG_COLOUR,
    LEFT_PANEL_WIDTH, 
    SCREEN_HEIGHT# - PANEL_MARGIN*2
)
panel.set_xy((LEFT_PANEL_WIDTH/2,  SCREEN_HEIGHT/2))

# event: scrolling
pysc.game['scrolling_offset'] = 0
def on_mouse_scroll(updown):
    if not panel.is_touching_mouse(): return
    if updown == 'up':
        pysc.game['scrolling_offset'] += 10
    else: 

        pysc.game['scrolling_offset'] -= 10
        
    pysc.game.broadcast_message("scrolling", pysc.game['scrolling_offset'])
    
pysc.game.when_mouse_scroll().add_handler(on_mouse_scroll)
pysc.game['left_panel'] = panel




import pyscratch as pysc
import pygame
from pyscratch import game

import right_paddle, left_paddle, ball, score_display

SCREEN_HEIGHT = 720
SCREEN_WIDTH = 1280


game['running']=False
game['score_left'] = 0
game['score_right'] = 0

game.load_sound('pong', 'assets/sound_effects/Metal Clang-SoundBible.com-19572601.wav')

font = pygame.font.SysFont(None, 48)
pysc.create_shared_data_display_sprite('left_score', font, size=(300, 60))

game.update_screen_mode((SCREEN_WIDTH, SCREEN_HEIGHT))
game.start(60, use_frame_time=True)



import numpy as np
import pyscratch as pysc
from pyscratch import game

game['centre_x'] = 1280/2
game['centre_h'] = 300
game['centre_d'] = 0
def tree_L():
    sp = pysc.create_animated_sprite("assets/used_by_examples/perspective_background/tree_L")  
    sp['d'] = pysc.random_number(100, 2000) 
    sp['world_x'] = pysc.random_number(0, 1000)
    sp.oob_limit=np.inf
    def on_game_start():
        sp.set_frame(int(pysc.random_number(0, 2.99)))
        sp.set_draggable(True)
        d = sp['d']-game['centre_d']
        game.change_layer(sp, 1000-int(sp['d']))
        while True:
            yield 1/game.framerate
            d = sp['d']-game['centre_d']

            new_scale = pysc.cap(300/d, 0.1, 3)
            if new_scale == 3:
                sp.hide()
            else:
                sp.show()
                sp.set_scale(new_scale)
            if d <= 0:
                sp.hide()

            sp.y = -200+game.screen_height-np.arctan(d/game['centre_h'])/np.pi*game.screen_height 
            sp.x = ((sp['world_x']-game['centre_x'])/d + 0.5)*game.screen_width

    sp.when_game_start().add_handler(on_game_start)

    return sp

for i in range(50):
    tree_L()



import pyscratch as pysc
from pyscratch.game_module import is_key_pressed


# my_sprite = pysc.create_animated_sprite("assets/run")

# def animate():
#     my_sprite.scale_by(5)
#     while True:
#         yield 0.2
#         my_sprite.next_frame()

# my_sprite.when_game_start().add_handler(animate)


my_sprite = pysc.create_animated_sprite("assets/my_sprite")
my_sprite.oob_limit=100

# variables for this sprite only
my_sprite['frame_interval'] = 0.4

# Event: on game start: switch frame
def animate():
    my_sprite.set_draggable(True)
    my_sprite.direction = 20
    my_sprite.scale_by(5)
    my_sprite.set_animation("idle") # reference the folder name of the animation
    while True:
        yield my_sprite['frame_interval']
        my_sprite.next_frame()

my_sprite.when_game_start().add_handler(animate)
#pysc.game._left_edge.hide()

# Event: on game start: movement
def movement(): 
    speed = 10
    
    #my_sprite.set_rotation_style_left_right()
    while True:
        yield 1/30
        my_sprite.if_on_edge_bounce()

        if pysc.is_key_pressed("d"):
            my_sprite.set_animation("run")
            #my_sprite.direction = 0
            my_sprite.x += speed
            my_sprite['frame_interval'] = 0.2

        elif pysc.is_key_pressed("a"):
            my_sprite.set_animation("run")
            #my_sprite.direction = 180
            my_sprite.x -= speed
            my_sprite['frame_interval'] = 0.2

        elif pysc.is_key_pressed("s"):
            my_sprite.set_animation("run")
            #my_sprite.direction = 180
            my_sprite.y += speed
            my_sprite['frame_interval'] = 0.2

        elif pysc.is_key_pressed("w"):
            my_sprite.set_animation("run")
            #my_sprite.direction = 180
            my_sprite.y -= speed
            my_sprite['frame_interval'] = 0.2

        elif pysc.is_key_pressed("space"):
            my_sprite.set_animation("run")
            #my_sprite.direction = 180
            my_sprite.move_indir(speed, -180)
            my_sprite['frame_interval'] = 0.2
        else: 
            my_sprite.set_animation("idle")
            my_sprite['frame_interval'] = 0.4

my_sprite.when_game_start().add_handler(movement)





import pyscratch as pysc
from pyscratch import game

def Enemy(position, speed, frame_idx, direction, bullet_period=0.4, path="assets/used_by_examples/bullet_hell/enemy"):
    enemy = pysc.create_animated_sprite(
        path,
        position=position
        )
    enemy.set_animation('normal')
    enemy.direction = direction#-90
    enemy.set_frame(frame_idx)
    enemy.set_scale(6/speed)


    def movement(_):
        while True: 
            yield 1/game.framerate

            enemy.move_indir(speed, 90)

    movement_event = enemy.when_timer_above(0).add_handler(movement)

    
    def explode_and_remove():
        movement_event.remove()
        enemy.set_animation('explosion')
        enemy.direction=0
        enemy.scale_by(1.3)
        enemy.when_timer_reset(0.1, 11).add_handler(lambda _: enemy.next_frame())
        enemy.when_timer_reset(0.1*11, 1).add_handler(lambda _: enemy.remove())

    def touch_player_bullet(_):
        player = game['player']

        while True: 
            yield 1/game.framerate
            if enemy.is_touching(player):
                explode_and_remove()
                return

            for b in game['player_bullets']:
                if enemy.is_touching(b):
                    b.remove()
                    explode_and_remove()
            
                    return

    enemy.when_timer_above(0).add_handler(touch_player_bullet)

    # shoot bullets
    StandardBullet = game['StandardBullet']
    enemy.when_timer_reset(bullet_period).add_handler(lambda _:  StandardBullet((enemy.x, enemy.y), direction+90, 10))

    return enemy


def spawn_standard_enemy():

    x = pysc.random_number(0, game.screen_width)
    y = 0
    frame_idx = int(pysc.random_number(0, 9.999))

    direction = pysc.random_number(-10, 10)

    Enemy((x,y), pysc.random_number(2, 6), frame_idx, direction)

def spawn_line():
    n = 6
    margin = 20
    itv = (game.screen_width-margin*2)//(n-1)
    y=0
    for i in range(n):

        x = itv*i+margin

        Enemy((x, y), 5, 0, 0, bullet_period=1)
    

def spawn_6_side_entry():

    pos1 = (0, 200)
    pos2 = (0, 400)
    pos3 = (0, 600)

    pos4 = (game.screen_width, 200)
    pos5 = (game.screen_width, 400)
    pos6 = (game.screen_width, 600)

    speed = 5
    fidx = 2

    Enemy(pos1, speed, fidx, 45-90)
    Enemy(pos2, speed, fidx, 45-90)
    Enemy(pos3, speed, fidx, 45-90)


    Enemy(pos4, speed, fidx, 45)
    Enemy(pos5, speed, fidx, 45)
    Enemy(pos6, speed, fidx, 45)


    

def spawn_kamikaze():
    x = pysc.random_number(0, game.screen_width)
    y = 0
    frame_idx = int(pysc.random_number(0, 9.999))
    direction = pysc.random_number(-10, 10)

    e = Enemy((x,y), 5, frame_idx, direction, bullet_period=1000)

    # move towards player
    e.when_timer_reset(1/game.framerate, int(10*game.framerate)).add_handler(lambda _: e.point_towards_sprite(game['player'], -90))

def spawn_laser_enemy(x, y):

    e = Enemy((x,y), speed=1, frame_idx=0, direction=0, bullet_period=1000, path="assets/used_by_examples/bullet_hell/enemy")
    e.set_animation('laser')
    e.scale_by(0.7)

    Laser = game['Laser']
    #player = game['player']
    e.when_timer_reset(3).add_handler(lambda _: Laser((e.x, e.y), (e.x, e.y+1000), 1))

def spawn_laser_enemy_pair(x):
    spawn_laser_enemy(x, 0)
    spawn_laser_enemy(game.screen_width-x, 0)



game.when_timer_reset(2).add_handler(lambda _:spawn_standard_enemy()) 

game.when_timer_above(3).add_handler(lambda _: spawn_line())
game.when_timer_above(4).add_handler(lambda _: spawn_6_side_entry())

game.when_timer_reset(7).add_handler(lambda _: spawn_line())
game.when_timer_reset(6).add_handler(lambda _: spawn_6_side_entry())

game.when_timer_reset(3).add_handler(lambda _: spawn_kamikaze())

game.when_timer_reset(5).add_handler(lambda _: game['ExplodingBullet']((720/2-pysc.random_number(-50, 50), 0), 90, 5))

game.when_timer_reset(10).add_handler(lambda _: spawn_laser_enemy(pysc.random_number(0, game.screen_width), 0))


#game.when_timer_reset(1).add_handler(lambda _: game['Laser']((0, 0), (game['player'].x, game['player'].y), 1))




        




import pyscratch as pysc
game = pysc.game

enemy = pysc.create_single_costume_sprite("assets/fish_red_skeleton_outline.png")
#enemy.set_draggable(True)


def clicked():
    """
    when the enemy is clicked: 
    change the enemy location
    """

    enemy.x = pysc.random_number(0, game.screen_width)
    enemy.y = pysc.random_number(0, game.screen_height)
    

enemy.when_this_sprite_clicked().add_handler(clicked)



import pyscratch as pysc
from pyscratch import game

SCREEN_HEIGHT = 720
SCREEN_WIDTH = 1280

paddle_colour = (200, 200, 200)
paddle_width = 20
paddle_height = 130
paddle_margin = 30

sprite = pysc.create_rect_sprite(paddle_colour, paddle_width, paddle_height,  position=(SCREEN_WIDTH-paddle_margin, SCREEN_HEIGHT//2))

#sprite.set_draggable(True)

game_start_event = sprite.when_game_start()
def movement():
    while True: 
        speed = 0
        
        if pysc.is_key_pressed('up'):
            speed -= 8
        if pysc.is_key_pressed('down'):
            speed += 8

        sprite.y += speed
        
        sprite.y = pysc.cap(sprite.y, 0+paddle_height//2, SCREEN_HEIGHT-paddle_height//2)


        # wait for 1 frame
        yield 1/game.framerate

game_start_event.add_handler(movement)


game['right_paddle'] = sprite


from pathlib import Path
from typing import Tuple
import pyscratch as pysc

from utils.render_wrapped_file_name import render_wrapped_file_name
from settings import *

width, height = 100, 140
colour = (130, 130, 130)

spacing = 10
n_cols = 5

def try_load_image(path):
    try: 
        return pysc.load_image(path)
    except:
        return None



def FileDisplay(path: Path, order: int, panel_top_left):
    sprite = pysc.create_rect_sprite(colour, width, height)
    
    # set the position

    panel_x, panel_y = panel_top_left
    sprite.x = spacing+(order%n_cols)*(width+spacing)+panel_x +width/2
    sprite.y = spacing+(order//n_cols)*(height+spacing)+panel_y +height/2

    # set the display

    surface = try_load_image(path)
    if not surface: 
        # TODO: messy af
        if path.is_file():
            sprite.remove()
            return None
        
        text = render_wrapped_file_name(path.name+'/', 8, DEFAULT_FONT24)
        sprite.draw(text, offset=(width/2, height/2) )
        #sprite.write_text(path.name+'/', DEFAULT_FONT24, offset=(width/2, height/2))
        
    else: 
        image_margin = 20
        text_height = 20
        fit = "horizontal" if surface.get_width() >= surface.get_height() else "vertical"
        surface = pysc.scale_to_fit_aspect(surface, (width-image_margin, height-text_height-image_margin ), fit)
        sprite.draw(surface, (width/2, (height-text_height)/2))
        
        text = render_wrapped_file_name(path.name, 8, DEFAULT_FONT24, color= (255, 255, 255), max_lines=2)
        sprite.draw(text,offset=(width/2, height-20), reset=False)



    sprite.sprite_data['is_file'] = path.is_file()


    def on_click():
        if not path.is_file():
            pysc.game.broadcast_message('folder_update', path)
        else:
            pysc.game.broadcast_message('image_selected', path)
            #pysc.game.broadcast_message('cut_or_nav_mode_change', 'cut')



    sprite.when_this_sprite_clicked().add_handler(on_click)


    def on_msg_mode_change(mode):
        if mode == 'nav':
            sprite.show()
            #pysc.game.bring_to_front(sprite)
        else:
            sprite.hide()

    sprite.when_receive_message('cut_or_nav_mode_change').add_handler(on_msg_mode_change)




    return sprite
    



import pyscratch as pysc
import pygame
game = pysc.game

enemy = pysc.create_single_costume_sprite("assets/fish_red_skeleton_outline.png")

# 4. Variable
game['clicked'] = False
game['score'] = 0
font = pygame.font.Font('assets/Kenney Future.ttf', size=18)
score_display = pysc.create_shared_data_display_sprite('score', font)
#score_display = pysc.create_shared_data_display_sprite('score')

# 3. Flow
# 4. Variable
def movement():
    """
    when_game_start: 
    the movement of the enemy
    """
    score_cool_down = 0
    while True:
        yield 1/game.framerate

        ## 4. Variable
        if (score_cool_down>0):
            score_cool_down -= 1/game.framerate

        if game['clicked']:
            continue 
    
        speed = max(1, (game['score']))
        centre = (game.screen_width/2, game.screen_height/2)

        ## 3. Flow Control
        mouse_x, mouse_y = pysc.get_mouse_pos()
        if enemy.distance_to((mouse_x, mouse_y)) < 200:
            enemy.point_towards_mouse()
            enemy.direction += 180
            enemy.move_indir(speed)
        
        else:
            enemy.point_towards(centre)
            enemy.move_indir(speed)

        ## 4. Variable
        if enemy.distance_to(centre) < 50 and (score_cool_down<=0):
            score_cool_down = .05
            game['score'] -= 1

enemy.when_game_start().add_handler(movement)
        

# 2. Basic Event
# 3. Flow
# 4. Variable
# 5. Sound 
def clicked():
    """
    when the enemy is clicked: 
    spin and reappear
    """

    ##
    game['score'] += 10
    game['clicked'] = True

    ##
    game.play_sound("hit")
    
    ## 3. Flow control (can be simplified to 2. Basic Event)
    t = 1
    for i in range(10):
        enemy.scale_by(0.9)
        enemy.direction += 30
        t -= 0.1
        enemy.set_transparency(t)
        yield 1/game.framerate


    enemy.set_scale(1)
    game['clicked'] = False

    # 2. Basic Event 
    new_x = pysc.random_number(0, game.screen_width)
    new_y = pysc.random_number(0, game.screen_height)

    enemy.x = new_x
    enemy.y = new_y
    enemy.set_transparency(1)
    
 
enemy.when_this_sprite_clicked().add_handler(clicked)


# 6. Backdrop
def game_end(): 
    """
    when_backdrop_switched: 
    hide the sprite 
    """
    enemy.hide()
    score_display.hide()

enemy.when_backdrop_switched(1).add_handler(game_end)




import pyscratch as pysc
from pyscratch import game
import chest, enemy, friend

# sound
game.load_sound("hit", "assets/impactMetal_light_004.ogg")
game.load_sound("background", "assets/Circus-Theme-Entry-of-the-Gladiators-Ragtime-Version(chosic.com).mp3")

def play_loop():
    """
    Continuously play the background music 
    """
    while True:
        game.play_sound("background", volume=0.1)
        yield 2*60+31 + 2 # or you can just put 153 
game.when_game_start().add_handler(play_loop)

# background
bg0 = pysc.load_image("assets/undersea_bg.png")
bg0 = pysc.scale_to_fit_aspect(bg0, (1024, 576))
game.add_backdrop('sea', bg0)

bg1 = pysc.load_image("assets/Cat In Space Wallpaper Hq.jpg")
bg1 = pysc.scale_to_fit_aspect(bg1, (1024, 576))
game.add_backdrop('lose', bg1)

# Background switches
game.switch_backdrop('sea')

def background():
    while True:
        yield 1/game.framerate
        if game['score'] < 0: 
            game.switch_backdrop('lose')
game.when_game_start().add_handler(background)


# starting the game
game.update_screen_mode((1024, 576))
game.start(show_mouse_position=True)


